"""Linting rule to ensure cog subclass names end with "Command(s)"."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from tokenize import TokenInfo
    from typing import Final

__all__: Sequence[str] = ("RuleCAR302",)


class RuleCAR302(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure cog subclass names end with "Command(s)"."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        needs_to_be_plural: object | None = ctx.get("needs_to_be_plural", None)
        if needs_to_be_plural is not None and not isinstance(needs_to_be_plural, bool):
            raise TypeError

        return (
            "Pycord cog subclass name should end with "
            f"{
                f"'Command{'s' if needs_to_be_plural else ''}Cog'"
                if needs_to_be_plural is not None
                else "'CommandCog' or 'CommandsCog'"
            }, "
            f"if that cog contains {
                f'{"multiple commands" if needs_to_be_plural else "a single command"}'
                if needs_to_be_plural is not None
                else 'command(s)'
            }"
        )

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    def _function_is_command(self, node: ast.AsyncFunctionDef) -> bool:
        if node.name.startswith("_") or node.name.startswith("autocomplete_"):
            return False

        if node.returns is not None and not (
            isinstance(node.returns, ast.Constant) and node.returns.value is None
        ):
            return False

        has_slash_command_decorator: bool = False

        decorator_node: ast.expr
        for decorator_node in node.decorator_list:
            if not isinstance(decorator_node, ast.Call):
                continue

            if utils.function_call_is_pycord_task_decorator(decorator_node):
                return False

            if utils.function_call_is_pycord_event_listener_decorator(decorator_node):
                return False

            if utils.function_call_is_pycord_slash_command_decorator(decorator_node):
                has_slash_command_decorator = True
                continue

            if utils.function_call_is_pycord_context_command_decorator(decorator_node):
                has_slash_command_decorator = True
                continue

            possible_slash_command_group_name: str
            match decorator_node.func:
                case ast.Attribute(
                    value=ast.Name(id=possible_slash_command_group_name),
                    attr="command",
                ):
                    if (
                        possible_slash_command_group_name
                        in self.plugin.found_slash_command_group_names
                    ):
                        has_slash_command_decorator = True
                        continue

        if has_slash_command_decorator:
            return has_slash_command_decorator

        second_argument: ast.arg
        match node.args:
            case ast.arguments(
                posonlyargs=[],
                args=[ast.arg(arg="self"), second_argument, *_],
            ):
                if second_argument.arg == "ctx":
                    return True

                if "context" in second_argument.arg.lower():
                    return True

                second_argument_annotation: utils.ASTNameID
                match second_argument.annotation:
                    case (
                        ast.Name(id=second_argument_annotation)
                        | ast.Constant(value=second_argument_annotation)
                    ):
                        if "context" in str(second_argument_annotation).lower():
                            return True

        return False

    @classmethod
    def _is_class_a_cog_subclass(
        cls, class_name: str, class_bases: Iterable[ast.expr]
    ) -> bool:
        if "cog" in class_name:
            return True

        base: ast.expr
        for base in class_bases:
            if isinstance(base, ast.Attribute) and "cog" in base.attr.lower():
                return True

            if isinstance(base, ast.Name) and "cog" in base.id.lower():
                return True

        return False

    @classmethod
    def _is_cog_class_mislabelled(
        cls, class_name: str, number_of_commands_in_class: int
    ) -> bool:
        if "base" in class_name.lower():
            return False

        if number_of_commands_in_class == 1 and not class_name.endswith("CommandCog"):
            return True

        return number_of_commands_in_class > 1 and not class_name.endswith("CommandsCog")

    @utils.generic_visit_before_return
    @override
    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        if not self._is_class_a_cog_subclass(node.name, node.bases):
            return

        number_of_commands_in_class: int = sum(
            (
                1
                for class_inner_node in node.body
                if (
                    isinstance(class_inner_node, ast.AsyncFunctionDef)
                    and self._function_is_command(class_inner_node)
                )
            ),
        )

        if number_of_commands_in_class < 0:
            NEGATIVE_COMMAND_FUNCTIONS_FOUND_MESSAGE: Final[str] = (
                "Found less than 0 command functions within cog class."
            )
            raise ValueError(NEGATIVE_COMMAND_FUNCTIONS_FOUND_MESSAGE)

        if number_of_commands_in_class == 0:
            return

        if self._is_cog_class_mislabelled(node.name, number_of_commands_in_class):
            self.problems[(node.lineno, node.col_offset + 6)] = {
                "needs_to_be_plural": number_of_commands_in_class > 1,
            }



"""Linting rule to ensure linting comments do not have an incorrect number of commas."""  # noqa: N999

import re
import tokenize
from collections.abc import Mapping
from enum import Enum
from typing import TYPE_CHECKING, override

from flake8_carrot.utils import CarrotRule, ProblemsContainer

if TYPE_CHECKING:
    import ast
    from collections.abc import Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR121",)


type _ErrorLocationContext = tuple[_IgnoreCommentType, bool]
type _ErrorLocationsMapping = Mapping[int, _ErrorLocationContext]
type _ErrorLocationsDict = dict[int, _ErrorLocationContext]


class _IgnoreCommentType(Enum):
    TYPE_IGNORE = (
        r"\s*#\s*type\s*:\s*ignore\s*\[\s*[a-z_-]+\s*(?:,\s*[a-z_-]+\s*)*((?:,\s*)+)\s*]",
        "`type: ignore`",
    )
    NOQA = (r"\s*#\s*noqa\s*:\s*[A-Z0-9]+(?:\s*,\s*[A-Z0-9]+)*\s*((?:,\s*)+)", "NOQA")


class RuleCAR121(CarrotRule):
    """Linting rule to ensure linting comments do not have an incorrect number of commas."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        ignore_comment_type: object | None = ctx.get("ignore_comment_type", None)
        if ignore_comment_type is not None and not isinstance(
            ignore_comment_type, _IgnoreCommentType
        ):
            raise TypeError

        multiple_commas: object | None = ctx.get("multiple_commas", None)
        if multiple_commas is not None and not isinstance(multiple_commas, bool):
            raise TypeError

        return (
            f"{ignore_comment_type.value[1] if ignore_comment_type is not None else 'Ignore'} "
            f"comment should not have {
                'trailing comma(s)'
                if multiple_commas is None
                else ('trailing commas' if multiple_commas else 'a trailing comma')
            }"
        )

    @classmethod
    def _get_single_error_locations(
        cls, ignore_comment_type: _IgnoreCommentType, line: str, offset: int = 0
    ) -> _ErrorLocationsDict:
        match: re.Match[str] | None = re.search(rf"{ignore_comment_type.value[0]}\Z", line)
        if match is None:
            return {}

        error_locations: _ErrorLocationsDict = {}

        match match.group(1).count(","):
            case 0:
                raise ValueError
            case 1:
                error_locations[offset + match.span(1)[0]] = ignore_comment_type, False
            case _:
                error_locations[offset + match.span(1)[0]] = ignore_comment_type, True

        return error_locations

    @classmethod
    def _get_type_ignore_first_error_locations(cls, line: str) -> _ErrorLocationsMapping:
        match: re.Match[str] | None = re.search(
            (
                rf"(?P<type_ignore>{
                    _IgnoreCommentType.TYPE_IGNORE.value[0].replace(
                        r'((?:,\s*)+)',
                        r'(?:,\s*)*',
                    )
                })"
                rf"(?P<noqa>{
                    _IgnoreCommentType.NOQA.value[0].replace(
                        r'((?:,\s*)+)',
                        r'(?:,\s*)*',
                    )
                })"
                r"\Z"
            ),
            line,
        )
        if match is None:
            return {}

        return cls._get_single_error_locations(
            _IgnoreCommentType.TYPE_IGNORE,
            match.group("type_ignore"),
            offset=match.span("type_ignore")[0],
        ) | cls._get_single_error_locations(
            _IgnoreCommentType.NOQA,
            match.group("noqa"),
            offset=match.span("noqa")[0],
        )

    @classmethod
    def _get_noqa_first_error_locations(cls, line: str) -> _ErrorLocationsMapping:
        match: re.Match[str] | None = re.search(
            (
                rf"(?P<noqa>{
                    _IgnoreCommentType.NOQA.value[0].replace(
                        r'((?:,\s*)+)',
                        r'(?:,\s*)*',
                    )
                })"
                rf"(?P<type_ignore>{
                    _IgnoreCommentType.TYPE_IGNORE.value[0].replace(
                        r'((?:,\s*)+)',
                        r'(?:,\s*)*',
                    )
                })"
                r"\Z"
            ),
            line,
        )
        if match is None:
            return {}

        return cls._get_single_error_locations(
            _IgnoreCommentType.NOQA,
            match.group("noqa"),
            offset=match.span("noqa")[0],
        ) | cls._get_single_error_locations(
            _IgnoreCommentType.TYPE_IGNORE,
            match.group("type_ignore"),
            offset=match.span("type_ignore")[0],
        )

    @classmethod
    def _get_all_error_locations(cls, line: str) -> _ErrorLocationsMapping:
        type_ignore_first_error_locations: _ErrorLocationsMapping = (
            cls._get_type_ignore_first_error_locations(line)
        )
        if type_ignore_first_error_locations:
            return type_ignore_first_error_locations

        noqa_first_error_locations: _ErrorLocationsMapping = (
            cls._get_noqa_first_error_locations(line)
        )
        if noqa_first_error_locations:
            return noqa_first_error_locations

        single_type_ignore_error_locations: _ErrorLocationsMapping = (
            cls._get_single_error_locations(
                _IgnoreCommentType.TYPE_IGNORE,
                line,
            )
        )
        if single_type_ignore_error_locations:
            return single_type_ignore_error_locations

        single_noqa_error_locations: _ErrorLocationsMapping = cls._get_single_error_locations(
            _IgnoreCommentType.NOQA,
            line,
        )
        if single_noqa_error_locations:
            return single_noqa_error_locations

        return {}

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.problems = ProblemsContainer(
            self.problems
            | {
                (file_token.start[0], file_token.start[1] + match_location): {
                    "ignore_comment_type": ignore_comment_type,
                    "multiple_commas": multiple_commas,
                }
                for file_token in file_tokens
                for match_location, (
                    ignore_comment_type,
                    multiple_commas,
                ) in self._get_all_error_locations(
                    file_token.string.rstrip(),
                ).items()
                if file_token.type == tokenize.COMMENT
            },
        )



"""Linting rule to warn about passing `*args` or `**kwargs` as function arguments."""  # noqa: N999

import ast
from enum import Enum
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR151",)


class RuleCAR151(CarrotRule, ast.NodeVisitor):
    """Linting rule to warn about passing `*args` or `**kwargs` as function arguments."""

    class _InvalidArgumentType(Enum):
        STAR_ARGS = "`*args`"
        STAR_STAR_KWARGS = "`**kwargs`"

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        invalid_argument_type: object | None = ctx.get("invalid_argument_type", None)
        if invalid_argument_type is not None and not isinstance(
            invalid_argument_type, cls._InvalidArgumentType
        ):
            raise TypeError

        return f"{
            invalid_argument_type.value
            if invalid_argument_type is not None
            else '`*args` or `**kwargs`'
        } passed to super-function call"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @utils.generic_visit_before_return
    @override
    def visit_Call(self, node: ast.Call) -> None:
        match node.func:
            case ast.Attribute(value=ast.Call(func=ast.Name(id="super"))):
                arg: ast.expr
                for arg in node.args:
                    if not isinstance(arg, ast.Starred):
                        continue

                    self.problems[(arg.lineno, arg.col_offset)] = {
                        "invalid_argument_type": self._InvalidArgumentType.STAR_ARGS,
                    }

                kwarg: ast.keyword
                for kwarg in node.keywords:
                    if kwarg.arg is not None:
                        continue

                    self.problems[(kwarg.lineno, kwarg.col_offset)] = {
                        "invalid_argument_type": self._InvalidArgumentType.STAR_STAR_KWARGS,
                    }

                return



"""Linting rule to ensure the `__all__` export is annotated as `Sequence[str]`."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR103",)


class RuleCAR103(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure the `__all__` export is annotated as `Sequence[str]`."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return "`__all__` export should be annotated as `Sequence[str]`"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @classmethod
    def _get_unannotated_all_from_assignment_targets(
        cls, targets: Iterable[ast.expr]
    ) -> ast.Name | None:
        target: ast.expr
        for target in targets:
            match target:
                case ast.Name(id="__all__"):
                    return target

        return None

    @utils.generic_visit_before_return
    @override
    def visit_Assign(self, node: ast.Assign) -> None:
        if self.plugin.first_all_export_line_numbers is None:
            return

        if node.lineno != self.plugin.first_all_export_line_numbers[0]:
            return

        all_assignment: ast.Name | None = self._get_unannotated_all_from_assignment_targets(
            node.targets,
        )
        if all_assignment is None:
            return

        self.problems.add_without_ctx(
            (
                all_assignment.lineno,
                (all_assignment.end_col_offset or all_assignment.col_offset),
            ),
        )

    @utils.generic_visit_before_return
    @override
    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        if self.plugin.first_all_export_line_numbers is None:
            return

        if node.lineno != self.plugin.first_all_export_line_numbers[0]:
            return

        match node.target:
            case ast.Name(id="__all__"):
                pass
            case _:
                return

        match node.annotation:
            case ast.Subscript(value=ast.Name(id="Sequence"), slice=ast.Name(id="str")):
                return
            case ast.Constant(value="Sequence[str]"):
                return
            case ast.Name(id="Sequence") | ast.Constant(value="Sequence"):
                self.problems.add_without_ctx(
                    (
                        node.annotation.lineno,
                        (node.annotation.end_col_offset or node.annotation.col_offset) - 1,
                    ),
                )
            case _:
                self.problems.add_without_ctx(
                    (node.annotation.lineno, node.annotation.col_offset),
                )



"""Linting rule to ensure pycord names don't have any invalid characters."""  # noqa: N999

import ast
import re
from enum import Enum
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo
    from typing import Final

__all__: Sequence[str] = ("RuleCAR301",)


class RuleCAR301(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure pycord names don't have any invalid characters."""

    class _FunctionType(Enum):
        COMMAND = "slash-command"
        OPTION = "option"

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        function_type: object | None = ctx.get("context_command_type", None)
        if function_type is not None and not isinstance(
            function_type, RuleCAR301._FunctionType
        ):
            raise TypeError

        incorrect_name: object | None = ctx.get("incorrect_name", None)
        if incorrect_name is not None:
            if not isinstance(incorrect_name, str):
                raise TypeError

            incorrect_name = incorrect_name.strip("\n\r\t '")

        invalid_character: object | None = ctx.get("invalid_character", None)
        if invalid_character is not None:
            if not isinstance(invalid_character, str):
                raise TypeError

            if len(invalid_character) != 1:
                MULTIPLE_INVALID_CHARACTERS_MESSAGE: Final[str] = (
                    "Multiple invalid characters were given."
                )
                raise ValueError(MULTIPLE_INVALID_CHARACTERS_MESSAGE)

        return f"Invalid character: {
            f"'{invalid_character}'" if invalid_character else ''
        } found within Pycord {
            function_type.value if function_type is not None else 'slash-command/option'
        } name{
            f" '{incorrect_name if len(incorrect_name) < 30 else f'{incorrect_name[:30]}...'}'"
            if incorrect_name
            else ''
        }"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    def _check_single_argument(
        self, argument: ast.expr, function_type: RuleCAR301._FunctionType
    ) -> None:
        if not isinstance(argument, ast.Constant):
            return

        if not argument.value or not isinstance(argument.value, str):
            return

        if function_type is self._FunctionType.COMMAND and argument.value.endswith(
            (".", "_", " ")
        ):
            COLUMN_OFFSET: Final[int] = (
                argument.end_col_offset - 1
                if argument.end_col_offset is not None
                else argument.col_offset + len(argument.value) - 1
            )
            self.problems[(argument.lineno, COLUMN_OFFSET)] = {
                "invalid_character": argument.value[-1],
                "function_type": function_type,
                "incorrect_name": argument.value,
            }

        for invalid_character in "`!¬£$€%^&*+=,<>?#~`":
            invalid_character_match: re.Match[str]
            for invalid_character_match in re.finditer(
                rf"\{invalid_character}", argument.value
            ):
                self.problems[
                    (
                        argument.lineno,
                        argument.col_offset + invalid_character_match.span()[0] + 1,
                    )
                ] = {
                    "invalid_character": invalid_character,
                    "function_type": function_type,
                    "incorrect_name": argument.value,
                }

    def _check_all_arguments(
        self, decorator_node: ast.Call, function_type: _FunctionType
    ) -> None:
        if decorator_node.args:
            self._check_single_argument(decorator_node.args[0], function_type)

        keyword_argument: ast.keyword
        for keyword_argument in decorator_node.keywords:
            if keyword_argument.arg == "name":
                self._check_single_argument(keyword_argument.value, function_type)
                return

    def _check_decorator(self, decorator_node: ast.expr) -> None:
        if not isinstance(decorator_node, ast.Call):
            return

        if utils.function_call_is_pycord_slash_command_decorator(decorator_node):
            self._check_all_arguments(decorator_node, self._FunctionType.COMMAND)
            return

        if utils.function_call_is_pycord_context_command_decorator(decorator_node):
            self._check_all_arguments(decorator_node, self._FunctionType.COMMAND)
            return

        if utils.function_call_is_pycord_option_decorator(decorator_node):
            self._check_all_arguments(decorator_node, self._FunctionType.OPTION)
            return

        possible_slash_command_group_name: str
        possible_pycord_decorator_name: str
        match decorator_node.func:
            case ast.Attribute(
                value=ast.Name(id=possible_slash_command_group_name),
                attr=possible_pycord_decorator_name,
            ):
                if (
                    possible_slash_command_group_name
                    in self.plugin.found_slash_command_group_names
                    and possible_pycord_decorator_name
                    in (
                        utils.PYCORD_SLASH_COMMAND_DECORATOR_NAMES
                        | utils.PYCORD_CONTEXT_COMMAND_DECORATOR_NAMES
                    )
                ):
                    self._check_all_arguments(decorator_node, self._FunctionType.COMMAND)
                    return

                if (
                    possible_slash_command_group_name
                    in self.plugin.found_slash_command_group_names
                    and possible_pycord_decorator_name in utils.PYCORD_OPTION_DECORATOR_NAMES
                ):
                    self._check_all_arguments(decorator_node, self._FunctionType.OPTION)
                    return

    @utils.generic_visit_before_return
    @override
    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        decorator_node: ast.expr
        for decorator_node in node.decorator_list:
            self._check_decorator(decorator_node)



"""Linting rule to ensure Pycord command and option names end with a full-stop."""  # noqa: N999

import ast
from enum import Enum
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR304",)


class RuleCAR304(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure Pycord command and option names end with a full-stop."""

    class _FunctionType(Enum):
        COMMAND = "command"
        OPTION = "option"

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        function_type: object | None = ctx.get("function_type", None)
        if function_type is not None and not isinstance(function_type, cls._FunctionType):
            raise TypeError

        return f"Pycord {
            function_type.value if function_type is not None else 'command/option'
        } description should end with a full-stop"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    def _check_single_argument(self, argument: ast.expr, function_type: _FunctionType) -> None:  # noqa: PLR0911, PLR0912
        if not isinstance(argument, ast.Constant):
            return

        if not argument.value or not isinstance(argument.value, str):
            return

        if argument.value.endswith("."):
            return
        if argument.value.endswith("!"):
            return
        if argument.value.endswith("?"):
            return
        if argument.value.endswith(";"):
            return
        if argument.value.endswith("%"):
            return
        if argument.value.endswith("]") and "[" in argument.value:
            return
        if argument.value.endswith(")") and "(" in argument.value:
            return
        if argument.value.endswith("}") and "{" in argument.value:
            return
        if argument.value.endswith(">") and "<" in argument.value:
            return
        if argument.value.endswith('"'):
            double_quote_count: int = argument.value.count('"')
            if double_quote_count > 1 and double_quote_count % 2 == 0:
                return
        if argument.value.endswith("'"):
            single_quote_count: int = argument.value.count("'")
            if single_quote_count > 1 and single_quote_count % 2 == 0:
                return
        if argument.value.endswith("`"):
            backtick_count: int = argument.value.count("`")
            if backtick_count > 1 and backtick_count % 2 == 0:
                return
        if argument.value.endswith("*"):
            asterisk_count: int = argument.value.count("*")
            if asterisk_count > 1 and asterisk_count % 2 == 0:
                return
        if argument.value.endswith("_"):
            underscore_count: int = argument.value.count("_")
            if underscore_count > 1 and underscore_count % 2 == 0:
                return
        if argument.value.endswith("~"):
            tilde_count: int = argument.value.count("~")
            if tilde_count > 1 and tilde_count % 2 == 0:
                return

        column_offset: int = (
            (argument.end_col_offset - 1) if argument.end_col_offset else argument.col_offset
        )
        self.problems[(argument.lineno, column_offset)] = {
            "function_type": function_type,
        }

    def _check_all_arguments(
        self, decorator_node: ast.Call, function_type: _FunctionType
    ) -> None:
        if len(decorator_node.args) >= 2:
            self._check_single_argument(decorator_node.args[1], function_type)

        keyword_argument: ast.keyword
        for keyword_argument in decorator_node.keywords:
            if keyword_argument.arg == "description":
                self._check_single_argument(keyword_argument.value, function_type)
                return

    def _check_decorator(self, decorator_node: ast.expr) -> None:
        if not isinstance(decorator_node, ast.Call):
            return

        if utils.function_call_is_pycord_slash_command_decorator(decorator_node):
            self._check_all_arguments(decorator_node, self._FunctionType.COMMAND)
            return

        if utils.function_call_is_pycord_context_command_decorator(decorator_node):
            self._check_all_arguments(decorator_node, self._FunctionType.COMMAND)
            return

        if utils.function_call_is_pycord_option_decorator(decorator_node):
            self._check_all_arguments(decorator_node, self._FunctionType.OPTION)
            return

        possible_slash_command_group_name: str
        possible_pycord_decorator_name: str
        match decorator_node.func:
            case ast.Attribute(
                value=ast.Name(id=possible_slash_command_group_name),
                attr=possible_pycord_decorator_name,
            ):
                if (
                    possible_slash_command_group_name
                    in self.plugin.found_slash_command_group_names
                    and possible_pycord_decorator_name
                    in (
                        utils.PYCORD_SLASH_COMMAND_DECORATOR_NAMES
                        | utils.PYCORD_CONTEXT_COMMAND_DECORATOR_NAMES
                    )
                ):
                    self._check_all_arguments(decorator_node, self._FunctionType.COMMAND)
                    return

                if (
                    possible_slash_command_group_name
                    in self.plugin.found_slash_command_group_names
                    and possible_pycord_decorator_name in utils.PYCORD_OPTION_DECORATOR_NAMES
                ):
                    self._check_all_arguments(decorator_node, self._FunctionType.OPTION)
                    return

    @utils.generic_visit_before_return
    @override
    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        decorator_node: ast.expr
        for decorator_node in node.decorator_list:
            self._check_decorator(decorator_node)



"""Linting rule to ensure ignore comments have the correct amount of whitespace."""  # noqa: N999

import re
import tokenize
from typing import TYPE_CHECKING, override

from flake8_carrot.utils import CarrotRule, ProblemsContainer

if TYPE_CHECKING:
    import ast
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo
    from typing import Final

__all__: Sequence[str] = ("RuleCAR120",)


class RuleCAR120(CarrotRule):
    """Linting rule to ensure ignore comments have the correct amount of whitespace."""

    TYPE_IGNORE_REGEX: Final[str] = (
        r"\s*#(\s*)type(\s*):(\s*)ignore(?:(\s*)\[(\s*)"
        r"[a-z_-]+(\s*)((?:,\s*[a-z_-]+\s*)*)(?:,(\s*))*])?"
    )
    NOQA_REGEX: Final[str] = (
        r"\s*#(\s*)noqa(?:(\s*):(\s*)[A-Z0-9]+"
        r"((?:\s*,\s*[A-Z0-9]+)*)(?:\s*,)*)?"
    )

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        replacement_message: object | None = ctx.get("replacement_message", None)
        if replacement_message is not None:
            if not isinstance(replacement_message, str):
                raise TypeError

            replacement_message = replacement_message.strip("\n\r\t '")

        return f"Incorrect amount of whitespace in ignore comment{
            f' ({replacement_message})' if replacement_message else ''
        }"

    @classmethod
    def _get_single_type_ignore_error_locations(
        cls, line: str, offset: int = 0
    ) -> dict[int, str]:
        match: re.Match[str] | None = re.search(rf"{cls.TYPE_IGNORE_REGEX}\Z", line)
        if match is None:
            return {}

        error_locations: dict[int, str] = {}

        if match.group(1) != " ":
            error_locations[offset + match.span(1)[0]] = "Replace with a single space"

        if match.group(2) != "":
            error_locations[offset + match.span(2)[0]] = "Remove all spaces"

        if match.group(3) != " ":
            error_locations[offset + match.span(3)[0]] = "Replace with a single space"

        group_number: int
        for group_number in (4, 5, 6):
            group: str | None = match.group(group_number)
            if group is None or group == "":
                continue
            error_locations[offset + match.span(group_number)[0]] = "Remove all spaces"

        group7: str | None = match.group(7)
        if group7 is not None:
            GROUP7_MATCH_START: Final[int] = match.span(7)[0]

            group7_match: re.Match[str]
            for group7_match in re.finditer(r",(\s*)[a-z_-]+(\s*)", group7):
                if group7_match.group(1) != " ":
                    error_locations[offset + GROUP7_MATCH_START + group7_match.span(1)[0]] = (
                        "Replace with a single space"
                    )
                if group7_match.group(2) != "":
                    error_locations[offset + GROUP7_MATCH_START + group7_match.span(2)[0]] = (
                        "Remove all spaces"
                    )

        trailing_comma_group: str | None = match.group(8)
        if trailing_comma_group is not None and trailing_comma_group != "":
            error_locations[offset + match.span(8)[0]] = "Remove all spaces"

        return error_locations

    @classmethod
    def _get_single_noqa_error_locations(cls, line: str, offset: int = 0) -> dict[int, str]:
        match: re.Match[str] | None = re.search(rf"{cls.NOQA_REGEX}\Z", line)
        if match is None:
            return {}

        error_locations: dict[int, str] = {}

        if match.group(1) != " ":
            error_locations[offset + match.span(1)[0]] = "Replace with a single space"

        group2: str | None = match.group(2)
        if group2 is not None and group2 != "":
            error_locations[offset + match.span(2)[0]] = "Remove all spaces"

        group3: str | None = match.group(3)
        if group3 is not None and group3 != " ":
            error_locations[offset + match.span(3)[0]] = "Replace with a single space"

        group4: str | None = match.group(4)
        if group4 is not None:
            GROUP4_MATCH_START: Final[int] = match.span(4)[0]

            group4_match: re.Match[str]
            for group4_match in re.finditer(r"(\s*),(\s*)[A-Z0-9]+", group4):
                if group4_match.group(1) != "":
                    error_locations[offset + GROUP4_MATCH_START + group4_match.span(1)[0]] = (
                        "Remove all spaces"
                    )
                if group4_match.group(2) != " ":
                    error_locations[offset + GROUP4_MATCH_START + group4_match.span(2)[0]] = (
                        "Replace with a single space"
                    )

        return error_locations

    @classmethod
    def _get_type_ignore_first_error_locations(cls, line: str) -> Mapping[int, str]:
        match: re.Match[str] | None = re.search(
            rf"(?P<type_ignore>{cls.TYPE_IGNORE_REGEX})(?P<noqa>{cls.NOQA_REGEX})\Z",
            line,
        )
        if match is None:
            return {}

        return cls._get_single_type_ignore_error_locations(
            match.group("type_ignore"),
            offset=match.span("type_ignore")[0],
        ) | cls._get_single_noqa_error_locations(
            match.group("noqa"),
            offset=match.span("noqa")[0],
        )

    @classmethod
    def _get_noqa_first_error_locations(cls, line: str) -> Mapping[int, str]:
        match: re.Match[str] | None = re.search(
            rf"(?P<noqa>{cls.NOQA_REGEX})(?P<type_ignore>{cls.TYPE_IGNORE_REGEX})\Z",
            line,
        )
        if match is None:
            return {}

        return cls._get_single_noqa_error_locations(
            match.group("noqa"),
            offset=match.span("noqa")[0],
        ) | cls._get_single_type_ignore_error_locations(
            match.group("type_ignore"),
            offset=match.span("type_ignore")[0],
        )

    @classmethod
    def _get_all_error_locations(cls, line: str) -> Mapping[int, str]:
        type_ignore_first_error_locations: Mapping[int, str] = (
            cls._get_type_ignore_first_error_locations(line)
        )
        if type_ignore_first_error_locations:
            return type_ignore_first_error_locations

        noqa_first_error_locations: Mapping[int, str] = cls._get_noqa_first_error_locations(
            line
        )
        if noqa_first_error_locations:
            return noqa_first_error_locations

        single_type_ignore_error_locations: Mapping[int, str] = (
            cls._get_single_type_ignore_error_locations(line)
        )
        if single_type_ignore_error_locations:
            return single_type_ignore_error_locations

        single_noqa_error_locations: Mapping[int, str] = cls._get_single_noqa_error_locations(
            line
        )
        if single_noqa_error_locations:
            return single_noqa_error_locations

        return {}

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.problems = ProblemsContainer(
            (
                self.problems
                | {
                    (file_token.start[0], file_token.start[1] + match_location): {
                        "replacement_message": replacement_message,
                    }
                    for file_token in file_tokens
                    for match_location, replacement_message in self._get_all_error_locations(
                        file_token.string.rstrip(),
                    ).items()
                    if file_token.type == tokenize.COMMENT
                }
            ),
        )



"""Linting rule to warn about uses of string functions that have will have no effect."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR141",)


class RuleCAR141(CarrotRule, ast.NodeVisitor):
    """Linting rule to warn about uses of string functions that have will have no effect."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return "String function seems to have no effect"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    def _check_for_string_function(self, statements: Iterable[ast.stmt]) -> None:
        statement: ast.stmt
        for statement in statements:
            match statement:
                case ast.Expr(
                    value=ast.Call(func=ast.Attribute(value=ast.Constant(value=str()))),
                ):
                    self.problems.add_without_ctx((statement.lineno, statement.col_offset))
                    continue
                case _:
                    continue

    @utils.generic_visit_before_return
    @override
    def visit_Module(self, node: ast.Module) -> None:
        self._check_for_string_function(node.body)

    @utils.generic_visit_before_return
    @override
    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self._check_for_string_function(node.body)

    @utils.generic_visit_before_return
    @override
    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self._check_for_string_function(node.body)

    @utils.generic_visit_before_return
    @override
    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self._check_for_string_function(node.body)

    @utils.generic_visit_before_return
    @override
    def visit_For(self, node: ast.For) -> None:
        self._check_for_string_function(node.body)
        self._check_for_string_function(node.orelse)

    @utils.generic_visit_before_return
    @override
    def visit_AsyncFor(self, node: ast.AsyncFor) -> None:
        self._check_for_string_function(node.body)
        self._check_for_string_function(node.orelse)

    @utils.generic_visit_before_return
    @override
    def visit_While(self, node: ast.While) -> None:
        self._check_for_string_function(node.body)
        self._check_for_string_function(node.orelse)

    @utils.generic_visit_before_return
    @override
    def visit_If(self, node: ast.If) -> None:
        self._check_for_string_function(node.body)
        self._check_for_string_function(node.orelse)

    @utils.generic_visit_before_return
    @override
    def visit_With(self, node: ast.With) -> None:
        self._check_for_string_function(node.body)

    @utils.generic_visit_before_return
    @override
    def visit_AsyncWith(self, node: ast.AsyncWith) -> None:
        self._check_for_string_function(node.body)

    @utils.generic_visit_before_return
    @override
    def visit_Try(self, node: ast.Try) -> None:
        self._check_for_string_function(node.body)
        self._check_for_string_function(node.orelse)
        self._check_for_string_function(node.finalbody)

    @utils.generic_visit_before_return
    @override
    def visit_TryStar(self, node: ast.TryStar) -> None:
        self._check_for_string_function(node.body)
        self._check_for_string_function(node.orelse)
        self._check_for_string_function(node.finalbody)

    @utils.generic_visit_before_return
    @override
    def visit_ExceptHandler(self, node: ast.ExceptHandler) -> None:
        self._check_for_string_function(node.body)

    @utils.generic_visit_before_return
    @override
    def visit_Match(self, node: ast.Match) -> None:
        case: ast.match_case
        for case in node.cases:
            self._check_for_string_function(case.body)



"""Utility/base functions and classes used throughout this project."""

import abc
import ast
import functools
import re
from collections.abc import Mapping
from typing import TYPE_CHECKING, final, override

from typed_classproperties import classproperty

if TYPE_CHECKING:
    from collections.abc import Callable, Collection, Generator, Iterable, Sequence
    from collections.abc import Set as AbstractSet
    from tokenize import TokenInfo
    from types import EllipsisType
    from typing import Final, Self

    from flake8_carrot.carrot import CarrotPlugin
    from flake8_carrot.tex_bot import TeXBotPlugin

__all__: Sequence[str] = (
    "ALL_PYCORD_FUNCTION_NAMES",
    "PYCORD_CONTEXT_COMMAND_DECORATOR_NAMES",
    "PYCORD_EVENT_LISTENER_DECORATOR_NAMES",
    "PYCORD_OPTION_DECORATOR_NAMES",
    "PYCORD_SLASH_COMMAND_DECORATOR_NAMES",
    "PYCORD_TASK_DECORATOR_NAMES",
    "ASTNameID",
    "BasePlugin",
    "BaseRule",
    "CarrotRule",
    "ProblemsContainer",
    "TeXBotRule",
    "function_call_is_any_pycord_decorator",
    "function_call_is_pycord_context_command_decorator",
    "function_call_is_pycord_event_listener_decorator",
    "function_call_is_pycord_option_decorator",
    "function_call_is_pycord_slash_command_decorator",
    "function_call_is_pycord_task_decorator",
    "generic_visit_before_return",
)


if __name__ == "__main__":
    CANNOT_RUN_AS_SCRIPT_MESSAGE: Final[str] = "This module cannot be run as a script."
    raise RuntimeError(CANNOT_RUN_AS_SCRIPT_MESSAGE)


if TYPE_CHECKING:
    type ASTNameID = str | bytes | int | float | complex | EllipsisType | None

    type _ProblemsContainerKey = tuple[int, int]
    type _ProblemsContainerValue = Mapping[str, object]
    type _ProblemsContainerMapping = Mapping[_ProblemsContainerKey, _ProblemsContainerValue]
    type _ProblemsContainerIterable = Iterable[
        tuple[_ProblemsContainerKey, _ProblemsContainerValue]
    ]


PPRINT_MODULES: Final[AbstractSet[str]] = {"astpretty"}
PYCORD_SLASH_COMMAND_DECORATOR_NAMES: Final[AbstractSet[str]] = {
    "application_command",
    "command",
    "slash_command",
    "ApplicationCommand",
    "SlashCommand",
    "SlashCommandGroup",
}
PYCORD_CONTEXT_COMMAND_DECORATOR_NAMES: Final[AbstractSet[str]] = {
    "user_command",
    "message_command",
    "UserCommand",
    "MessageCommand",
}
PYCORD_OPTION_DECORATOR_NAMES: Final[AbstractSet[str]] = {
    "option",
    "Option",
    "ThreadOption",
    "OptionChoice",
}
PYCORD_TASK_DECORATOR_NAMES: Final[AbstractSet[str]] = {"loop", "Loop", "SleepHandle"}
PYCORD_EVENT_LISTENER_DECORATOR_NAMES: Final[AbstractSet[str]] = {"listen", "listener"}
ALL_PYCORD_FUNCTION_NAMES: Final[AbstractSet[str]] = (
    PYCORD_SLASH_COMMAND_DECORATOR_NAMES
    | PYCORD_CONTEXT_COMMAND_DECORATOR_NAMES
    | PYCORD_OPTION_DECORATOR_NAMES
    | PYCORD_TASK_DECORATOR_NAMES
    | PYCORD_EVENT_LISTENER_DECORATOR_NAMES
)


class BasePlugin(abc.ABC):
    """Base plugin class to hold a selection of linting rules."""

    @classproperty
    @abc.abstractmethod
    def RULES(cls) -> Collection[type[BaseRule[Self]]]:  # noqa: D102, N802
        pass

    @override
    def __init__(
        self,
        tree: ast.AST,
        file_tokens: "Sequence[TokenInfo]",  # noqa: UP037
        lines: "Sequence[str]",  # noqa: UP037
    ) -> None:
        if isinstance(tree, ast.Module):
            pass
        elif isinstance(tree, ast.Interactive):
            tree = ast.Module(body=tree.body, type_ignores=[])
        elif isinstance(tree, ast.Expression):
            tree = ast.Module(body=[ast.Expr(tree.body)], type_ignores=[])
        else:
            CANNOT_RUN_WITH_NON_MODULE_MESSAGE: Final[str] = (
                "Cannot run flake8-carrot plugin with tree that is not an 'ast.Module'."
            )
            raise TypeError(CANNOT_RUN_WITH_NON_MODULE_MESSAGE)

        self._tree: ast.Module = tree
        self._file_tokens: Sequence[TokenInfo] = file_tokens
        self._lines: Sequence[str] = lines

    def run(self) -> Generator[tuple[int, int, str, type[Self]]]:
        """Perform complete linting over the stored Flake8 code context."""
        RuleClass: type[BaseRule[Self]]
        for RuleClass in self.RULES:
            rule: BaseRule[Self] = RuleClass(plugin=self)
            rule.run_check(tree=self._tree, file_tokens=self._file_tokens, lines=self._lines)

            line_number: int
            column_number: int
            ctx: Mapping[str, object]
            for (line_number, column_number), ctx in rule.problems.items():
                yield line_number, column_number, rule.format_error_message(ctx), type(self)


class ProblemsContainer(dict["_ProblemsContainerKey", "_ProblemsContainerValue"]):
    """Collection of problem locations mapped to error message contexts."""

    @classmethod
    def clean_key(cls, key: _ProblemsContainerKey | str) -> _ProblemsContainerKey:
        """Ensure the given problem location matches the required format."""
        if isinstance(key, str):
            match: re.Match[str] | None = re.fullmatch(
                r"\A(?P<line_number>\d+),(?P<column_number>\d+)\Z",
                key,
            )
            if match is None:
                INVALID_PROBLEM_LOCATION_MESSAGE: Final[str] = (
                    f"Invalid problem location: `{key}`."
                )
                raise ValueError(INVALID_PROBLEM_LOCATION_MESSAGE)

            key = (int(match.group("line_number")), int(match.group("column_number")))

        if key[0] < 1 or key[1] < 0:
            NEGATIVE_PROBLEM_LOCATION_MESSAGE: Final[str] = (
                "Problem locations cannot be negative."
            )
            raise ValueError(NEGATIVE_PROBLEM_LOCATION_MESSAGE)

        return key

    @override
    def __init__(
        self,
        mapping: _ProblemsContainerMapping | _ProblemsContainerIterable | None = None,
        /,
        **kwargs: _ProblemsContainerValue,  # noqa: CAR150
    ) -> None:
        mapping = (
            {}
            if mapping is None
            else {
                self.clean_key(key): value
                for key, value in (
                    mapping.items() if isinstance(mapping, Mapping) else mapping
                )
            }
        )

        if kwargs:
            mapping.update({self.clean_key(key): value for key, value in kwargs.items()})

        super().__init__(mapping)

    @override
    def __setitem__(
        self, key: _ProblemsContainerKey, value: _ProblemsContainerValue, /
    ) -> None:
        super().__setitem__(self.clean_key(key), value)

    def add_without_ctx(self, problem_location: _ProblemsContainerKey) -> None:
        """Add a problem at the given location without a context."""
        self[problem_location] = {}


class BaseRule[T_plugin: BasePlugin](abc.ABC):
    """Base rule class defining common plugin-based functionality."""

    @override
    def __init__(self, plugin: T_plugin) -> None:
        self.plugin: T_plugin = plugin
        self.problems: ProblemsContainer = ProblemsContainer()

        super().__init__()

    @abc.abstractmethod
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        """Update the problems-dict from those arising from the given code."""

    @classmethod
    @abc.abstractmethod
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        pass

    @classmethod
    @final
    def format_error_message(cls, ctx: Mapping[str, object]) -> str:
        """Retrieve the formatted error message for this rule with the given context."""
        return (
            f"{cls.__name__.lower().removeprefix('rule').upper()} "
            f"{cls._format_error_message(ctx)}"
        )


class CarrotRule(BaseRule["CarrotPlugin"], abc.ABC):
    """Base rule class for all "CAR" lint rules."""

    @override
    def __init__(self, plugin: CarrotPlugin) -> None:
        super().__init__(plugin)


class TeXBotRule(BaseRule["TeXBotPlugin"], abc.ABC):
    """Base rule class for all "TXB" lint rules."""

    @override
    def __init__(self, plugin: TeXBotPlugin) -> None:
        super().__init__(plugin)


def _function_call_is_pycord_function_from_commands_module(
    node: ast.Call, decorator_names: Collection[str]
) -> bool:
    function_name: str
    match node.func:
        case (
            ast.Name(id=function_name)
            | ast.Attribute(
                value=(
                    ast.Name(id="discord")
                    | ast.Attribute(
                        value=ast.Name(id="discord"),
                        attr="commands",
                    )
                    | ast.Attribute(
                        value=ast.Attribute(
                            value=ast.Name(id="discord"),
                            attr="commands",
                        ),
                        attr=("core" | "options"),
                    )
                ),
                attr=function_name,
            )
        ):
            return function_name in decorator_names

        case _:
            return False


def function_call_is_pycord_slash_command_decorator(node: ast.Call) -> bool:
    """Check if the given call AST node is calling the pycord slash-command decorator."""
    return _function_call_is_pycord_function_from_commands_module(
        node,
        PYCORD_SLASH_COMMAND_DECORATOR_NAMES,
    )


def function_call_is_pycord_context_command_decorator(node: ast.Call) -> bool:
    """Check if the given call AST node is calling the pycord context-command decorator."""
    return _function_call_is_pycord_function_from_commands_module(
        node,
        PYCORD_CONTEXT_COMMAND_DECORATOR_NAMES,
    )


def function_call_is_pycord_option_decorator(node: ast.Call) -> bool:
    """Check if the given call AST node is calling the pycord command option decorator."""
    return _function_call_is_pycord_function_from_commands_module(
        node,
        PYCORD_OPTION_DECORATOR_NAMES,
    )


def function_call_is_pycord_task_decorator(node: ast.Call) -> bool:
    """Check if the given function call AST node is calling the pycord task decorator."""
    function_name: str
    match node.func:
        case (
            ast.Name(id=function_name)
            | ast.Attribute(
                value=(
                    ast.Attribute(
                        value=(
                            ast.Name(id="discord")
                            | ast.Attribute(value=ast.Name(id="discord"), attr="ext")
                        ),
                        attr="tasks",
                    )
                ),
                attr=function_name,
            )
        ):
            return function_name in PYCORD_TASK_DECORATOR_NAMES

    module_name: str
    match node.func:
        case ast.Attribute(
            value=ast.Name(id=module_name),
            attr=function_name,
        ):
            if function_name not in PYCORD_TASK_DECORATOR_NAMES:
                return False

            if "bot" in module_name.lower() or "client" in module_name.lower():
                return True

            return module_name in ("discord", "tasks")

    return False


def function_call_is_pycord_event_listener_decorator(node: ast.Call) -> bool:
    """Check if the given call AST node is calling a pycord event listener decorator."""
    function_name: str
    match node.func:
        case (
            ast.Name(id=function_name)
            | ast.Attribute(
                value=ast.Attribute(
                    value=ast.Attribute(
                        value=ast.Name(id="discord"),
                        attr="ext",
                    ),
                    attr="commands",
                ),
                attr=function_name,
            )
        ):
            return function_name in PYCORD_EVENT_LISTENER_DECORATOR_NAMES

    module_name: str
    match node.func:
        case ast.Attribute(value=ast.Name(id=module_name), attr=function_name):
            if function_name not in PYCORD_EVENT_LISTENER_DECORATOR_NAMES:
                return False

            if any(value in module_name.lower() for value in ("bot", "client", "cog")):
                return True

            return module_name in ("discord", "commands")

        case ast.Attribute(
            value=ast.Attribute(value=ast.Name(id="discord"), attr=module_name),
            attr=function_name,
        ):
            if function_name not in PYCORD_EVENT_LISTENER_DECORATOR_NAMES:
                return False

            if any(value in module_name.lower() for value in ("bot", "client", "cog")):
                return True

            return module_name == "commands"

    return False


def function_call_is_any_pycord_decorator(node: ast.Call) -> bool:
    """Check if the given function call AST node is calling a pycord decorator."""
    return any(
        (
            function_call_is_pycord_slash_command_decorator(node),
            function_call_is_pycord_context_command_decorator(node),
            function_call_is_pycord_option_decorator(node),
            function_call_is_pycord_task_decorator(node),
            function_call_is_pycord_event_listener_decorator(node),
        ),
    )


def generic_visit_before_return[T_Visitor: ast.NodeVisitor, T_Node: ast.AST](
    func: Callable[[T_Visitor, T_Node], None],
) -> Callable[[T_Visitor, T_Node], None]:
    """Ensure a NodeVisitor overriden method calls generic_visit() before returning."""

    @functools.wraps(func)
    def wrapper(self: T_Visitor, node: T_Node) -> None:
        func(self, node)
        self.generic_visit(node)

    return wrapper



"""Linting rule to ensure union typesin `isintance()` calls are replaced by tuples."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR170",)


class RuleCAR170(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure union typesin `isintance()` calls are replaced by tuples."""

    @classmethod
    def _convert_expr_to_tuple(cls, expr: ast.expr) -> str:
        match expr:
            case ast.BinOp(op=ast.BitOr(), left=left, right=right):
                return (
                    f"{cls._convert_expr_to_tuple(left)}, {cls._convert_expr_to_tuple(right)}"
                )
            case ast.BinOp() | ast.BoolOp():
                return f"({ast.unparse(expr)})"
            case _:
                return ast.unparse(expr)

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        type_union: object | None = ctx.get("type_union", None)
        type_tuple: str = ""
        match type_union:
            case None:
                pass
            case ast.BinOp(op=ast.BitOr(), left=left, right=right):
                type_union = ast.unparse(type_union)
                type_tuple = f"({cls._convert_expr_to_tuple(left)}, {
                    cls._convert_expr_to_tuple(right)
                })"
            case _:
                raise TypeError

        return f"Replace type union{
            f' `{type_union if len(type_union) < 30 else f"{type_union[:30]}..."}`'
            if type_union
            else ''
        } with tuple{
            f' `{type_tuple if len(type_tuple) < 30 else f"{type_tuple[:30]}..."}`'
            if type_tuple
            else ''
        }"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @override
    def generic_visit(self, node: ast.AST) -> None:
        match node:
            case ast.Call(
                func=(
                    ast.Name(id="isinstance")
                    | ast.Name(id="issubclass")
                    | ast.Attribute(
                        value=ast.Name(id="builtins"), attr="isinstance" | "issubclass"
                    )
                ),
                args=[_, ast.BinOp(op=ast.BitOr()) as type_union],
            ):
                self.problems[(node.lineno, node.col_offset)] = {"type_union": type_union}

        super().generic_visit(node=node)



"""
Linting rule to ensure the `__all__` export is a tuple.

Only applies to simple static `__all__` exports.
"""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR104",)


class RuleCAR104(CarrotRule, ast.NodeVisitor):
    """
    Linting rule to ensure the `__all__` export is a tuple.

    Only applies to simple static `__all__` exports.
    """

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return "Simple `__all__` export should be of type `tuple`, not `list`"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @classmethod
    def _targets_contain_all(cls, targets: Iterable[ast.expr]) -> bool:
        target: ast.expr
        for target in targets:
            match target:
                case ast.Name(id="__all__"):
                    return True

        return False

    @utils.generic_visit_before_return
    @override
    def visit_Assign(self, node: ast.Assign) -> None:
        if self.plugin.first_all_export_line_numbers is None:
            return

        if node.lineno != self.plugin.first_all_export_line_numbers[0]:
            return

        if not isinstance(node.value, ast.List):
            return

        if self._targets_contain_all(node.targets):
            self.problems.add_without_ctx((node.value.lineno, node.value.col_offset))

    @utils.generic_visit_before_return
    @override
    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        if self.plugin.first_all_export_line_numbers is None:
            return

        if node.lineno != self.plugin.first_all_export_line_numbers[0]:
            return

        if not isinstance(node.value, ast.List):
            return

        match node.target:
            case ast.Name(id="__all__"):
                self.problems.add_without_ctx((node.value.lineno, node.value.col_offset))



"""Linting rules under the "CAR" category."""

import ast
from typing import TYPE_CHECKING, override

from typed_classproperties import classproperty

from flake8_carrot import utils
from flake8_carrot.utils import BasePlugin, CarrotRule

from .CAR101 import RuleCAR101
from .CAR102 import RuleCAR102
from .CAR103 import RuleCAR103
from .CAR104 import RuleCAR104
from .CAR105 import RuleCAR105
from .CAR110 import RuleCAR110
from .CAR111 import RuleCAR111
from .CAR120 import RuleCAR120
from .CAR121 import RuleCAR121
from .CAR122 import RuleCAR122
from .CAR123 import RuleCAR123
from .CAR124 import RuleCAR124
from .CAR140 import RuleCAR140
from .CAR141 import RuleCAR141
from .CAR150 import RuleCAR150
from .CAR151 import RuleCAR151
from .CAR160 import RuleCAR160
from .CAR161 import RuleCAR161
from .CAR162 import RuleCAR162
from .CAR163 import RuleCAR163
from .CAR170 import RuleCAR170
from .CAR180 import RuleCAR180
from .CAR201 import RuleCAR201
from .CAR202 import RuleCAR202
from .CAR301 import RuleCAR301
from .CAR302 import RuleCAR302
from .CAR303 import RuleCAR303
from .CAR304 import RuleCAR304
from .CAR305 import RuleCAR305
from .CAR401 import RuleCAR401
from .CAR501 import RuleCAR501
from .CAR601 import RuleCAR601
from .CAR602 import RuleCAR602
from .CAR610 import RuleCAR610

if TYPE_CHECKING:
    from collections.abc import Collection, Sequence
    from collections.abc import Set as AbstractSet
    from tokenize import TokenInfo

__all__: Sequence[str] = (
    "CarrotPlugin",
    "CarrotRule",
    "RuleCAR101",
    "RuleCAR102",
    "RuleCAR103",
    "RuleCAR104",
    "RuleCAR105",
    "RuleCAR110",
    "RuleCAR111",
    "RuleCAR120",
    "RuleCAR121",
    "RuleCAR122",
    "RuleCAR123",
    "RuleCAR124",
    "RuleCAR140",
    "RuleCAR141",
    "RuleCAR150",
    "RuleCAR151",
    "RuleCAR160",
    "RuleCAR161",
    "RuleCAR162",
    "RuleCAR163",
    "RuleCAR170",
    "RuleCAR180",
    "RuleCAR201",
    "RuleCAR202",
    "RuleCAR301",
    "RuleCAR302",
    "RuleCAR303",
    "RuleCAR304",
    "RuleCAR305",
    "RuleCAR401",
    "RuleCAR501",
    "RuleCAR601",
    "RuleCAR602",
    "RuleCAR610",
)


class _ContextValuesFinder(ast.NodeVisitor):
    @override
    def __init__(self) -> None:
        self.found_slash_command_group_names: set[str] = set()
        self.first_all_export_line_numbers: tuple[int, int] | None = None
        self.pprint_imported_for_debugging: bool = False

        # TODO: Implement finding  # noqa: FIX002
        self.found_loggers: set[ast.Assign | ast.AnnAssign] = set()

    @classmethod
    def _node_is_slash_command_group_assignment(cls, node: ast.Assign | ast.AnnAssign) -> bool:
        match node.value:
            case ast.Call(
                func=(
                    ast.Name(id="SlashCommandGroup")
                    | ast.Attribute(
                        value=(
                            ast.Name(id="discord")
                            | ast.Attribute(
                                value=ast.Name(id="discord"),
                                attr="commands",
                            )
                            | ast.Attribute(
                                value=ast.Attribute(
                                    value=ast.Name(id="discord"),
                                    attr="commands",
                                ),
                                attr="core",
                            )
                        ),
                        attr="SlashCommandGroup",
                    )
                ),
            ):
                return True

            case _:
                return False

    @utils.generic_visit_before_return
    @override
    def visit_Assign(self, node: ast.Assign) -> None:
        if self._node_is_slash_command_group_assignment(node):
            target: ast.expr
            for target in node.targets:
                if not isinstance(target, ast.Name):
                    continue

                self.found_slash_command_group_names.add(target.id)

        if self.first_all_export_line_numbers is None:
            for target in node.targets:
                match target:
                    case ast.Name(id="__all__"):
                        self.first_all_export_line_numbers = (
                            node.lineno,
                            node.end_lineno or node.lineno,
                        )
                        break

        # TODO: Find loggers using rule 201  # noqa: FIX002

    @utils.generic_visit_before_return
    @override
    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        if isinstance(node.target, ast.Name) and self._node_is_slash_command_group_assignment(
            node
        ):
            self.found_slash_command_group_names.add(node.target.id)

        match node.target:
            case ast.Name(id="__all__") if self.first_all_export_line_numbers is None:
                self.first_all_export_line_numbers = (
                    node.lineno,
                    (node.end_lineno or node.lineno),
                )

    @utils.generic_visit_before_return
    @override
    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        if node.module in utils.PPRINT_MODULES:
            import_alias: ast.alias
            for import_alias in node.names:
                if "print" in import_alias.name:
                    self.pprint_imported_for_debugging = True
                    break


class CarrotPlugin(BasePlugin):
    """Plugin class holding all "TXB" rules to be run on some code provided by Flake8."""

    @classproperty
    @override
    def RULES(cls) -> Collection[type[CarrotRule]]:
        return {
            RuleCAR101,
            RuleCAR102,
            RuleCAR103,
            RuleCAR104,
            RuleCAR105,
            RuleCAR110,
            RuleCAR111,
            RuleCAR120,
            RuleCAR121,
            RuleCAR122,
            RuleCAR123,
            RuleCAR124,
            RuleCAR140,
            RuleCAR141,
            RuleCAR150,
            RuleCAR151,
            RuleCAR160,
            RuleCAR161,
            RuleCAR162,
            RuleCAR163,
            RuleCAR170,
            RuleCAR180,
            RuleCAR201,
            RuleCAR202,
            RuleCAR301,
            RuleCAR302,
            RuleCAR303,
            RuleCAR304,
            RuleCAR305,
            RuleCAR401,
            RuleCAR501,
            RuleCAR601,
            RuleCAR602,
            RuleCAR610,
        }

    @override
    def __init__(
        self,
        tree: ast.AST,
        file_tokens: "Sequence[TokenInfo]",  # noqa: UP037
        lines: "Sequence[str]",  # noqa: UP037
    ) -> None:
        context_values_finder: _ContextValuesFinder = _ContextValuesFinder()
        context_values_finder.visit(tree)

        self._found_slash_command_group_names: AbstractSet[str] = (
            context_values_finder.found_slash_command_group_names
        )
        self._first_all_export_line_numbers: tuple[int, int] | None = (
            context_values_finder.first_all_export_line_numbers
        )
        self._pprint_imported_for_debugging: bool = (
            context_values_finder.pprint_imported_for_debugging
        )
        self._found_loggers: AbstractSet[ast.Assign | ast.AnnAssign] = (
            context_values_finder.found_loggers
        )

        super().__init__(tree=tree, file_tokens=file_tokens, lines=lines)

    @property
    def found_slash_command_group_names(self) -> AbstractSet[str]:  # noqa: D102
        return self._found_slash_command_group_names

    @property
    def first_all_export_line_numbers(self) -> tuple[int, int] | None:  # noqa: D102
        return self._first_all_export_line_numbers

    @property
    def pprint_imported_for_debugging(self) -> bool:  # noqa: D102
        return self._pprint_imported_for_debugging

    @property
    def found_loggers(self) -> AbstractSet[ast.Assign | ast.AnnAssign]:  # noqa: D102
        return self._found_loggers



import ast
import tokenize
from io import StringIO
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence
    from collections.abc import Set as AbstractSet

    from flake8_carrot.utils import BasePlugin

__all__: Sequence[str] = ("apply_plugin_to_ast",)


def apply_plugin_to_ast(
    raw_testing_ast: str, plugin_class: type[BasePlugin]
) -> AbstractSet[str]:
    """Retrieve all the warnings from applying all of a plugin's rules to an AST."""
    converted_lines: Sequence[str] = raw_testing_ast.split("\n")
    if not converted_lines[-1]:
        converted_lines = [f"{line}\n" for line in converted_lines[:-1]]
    else:
        converted_lines = [f"{line}\n" for line in converted_lines[:-1]] + [
            converted_lines[-1]
        ]

    return {
        f"{line}:{column + 1} {message}"
        for line, column, message, _ in plugin_class(
            tree=ast.parse(raw_testing_ast),
            file_tokens=list(tokenize.generate_tokens(StringIO(raw_testing_ast).readline)),
            lines=converted_lines,
        ).run()
    }



"""Linting rule to ensure a double newline is present after the `__all__` export."""  # noqa: N999

from typing import TYPE_CHECKING, override

from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    import ast
    from collections.abc import Iterator, Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR110",)


class RuleCAR110(CarrotRule):
    """Linting rule to ensure a double newline is present after the `__all__` export."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return "Double newline is required after `__all__` export"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        if self.plugin.first_all_export_line_numbers is None:
            return

        if not lines[self.plugin.first_all_export_line_numbers[1] - 1].endswith("\n"):
            return

        remaining_lines: Iterator[str] = iter(
            lines[self.plugin.first_all_export_line_numbers[1] :],
        )

        first_line_after: str | None = next(remaining_lines, None)
        if first_line_after is None:
            return

        if first_line_after.strip("\n\r"):
            self.problems.add_without_ctx(
                (self.plugin.first_all_export_line_numbers[1] + 1, 0),
            )
            return

        second_line_after: str | None = next(remaining_lines, None)
        if second_line_after is None:
            return

        if second_line_after.strip("\n\r"):
            self.problems.add_without_ctx(
                (self.plugin.first_all_export_line_numbers[1] + 1, 0),
            )
            return

        third_line_after: str | None = next(remaining_lines, None)
        if third_line_after is None or third_line_after.strip("\n\r"):
            return

        self.problems.add_without_ctx(
            (self.plugin.first_all_export_line_numbers[1] + 3, 0),
        )



"""Linting rule to ensure Pycord command and option names are in the correct format."""  # noqa: N999

import ast
import re
from enum import Enum
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo
    from typing import Literal

__all__: Sequence[str] = ("RuleCAR303",)


class RuleCAR303(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure Pycord command and option names are in the correct format."""

    class _FunctionType(Enum):
        COMMAND = "slash-command"
        OPTION = "option"

    class _InvalidArgumentReason(Enum):
        REQUIRES_HYPHENATION = "should be hyphenated"
        REQUIRES_LOWERCASING = "should be lowercased"
        REQUIRES_BOTH = "should be both hyphenated and lowercased"

        @classmethod
        def from_bools(
            cls, *, requires_hyphenation: bool, requires_lowercasing: bool
        ) -> RuleCAR303._InvalidArgumentReason | Literal[False]:
            """Generate the correct enum type from separate boolean flags."""
            if requires_hyphenation and requires_lowercasing:
                return cls.REQUIRES_BOTH

            if requires_hyphenation:
                return cls.REQUIRES_HYPHENATION

            if requires_lowercasing:
                return cls.REQUIRES_LOWERCASING

            return False

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        function_type: object | None = ctx.get("function_type", None)
        if function_type is not None and not isinstance(function_type, cls._FunctionType):
            raise TypeError

        incorrect_name: object | None = ctx.get("incorrect_name", None)
        if incorrect_name is not None:
            if not isinstance(incorrect_name, str):
                raise TypeError

            incorrect_name = incorrect_name.strip("\n\r\t '")

        invalid_argument_reason: object | None = ctx.get("invalid_argument_reason", None)
        if invalid_argument_reason is not None and not isinstance(
            invalid_argument_reason, cls._InvalidArgumentReason
        ):
            raise TypeError

        corrected_name: str | None = (
            f"{
                incorrect_name[:-1]
                .replace(
                    '.',
                    '-',
                )
                .replace(
                    ' ',
                    '-',
                )
                .replace(
                    '_',
                    '-',
                )
            }{incorrect_name[-1]}"
            if (
                incorrect_name
                and (
                    invalid_argument_reason is cls._InvalidArgumentReason.REQUIRES_HYPHENATION
                    or invalid_argument_reason is cls._InvalidArgumentReason.REQUIRES_BOTH
                )
            )
            else incorrect_name
        )

        corrected_name = corrected_name.lower() if corrected_name else corrected_name

        return f"Pycord {
            function_type.value if function_type is not None else 'slash-command/option'
        } name{
            f" '{incorrect_name if len(incorrect_name) < 30 else f'{incorrect_name[:30]}...'}'"
            if incorrect_name
            else ''
        } {
            invalid_argument_reason.value
            if invalid_argument_reason is not None
            else 'should be hyphenated and/or lowercased'
        } {
            f": '{
                corrected_name if len(corrected_name) < 30 else f'{corrected_name[:30]}...'
            }'"
            if corrected_name
            else ''
        }"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    def _check_single_argument(self, argument: ast.expr, function_type: _FunctionType) -> None:
        if not isinstance(argument, ast.Constant):
            return

        if not argument.value or not isinstance(argument.value, str):
            return

        reason: RuleCAR303._InvalidArgumentReason | Literal[False] = (
            self._InvalidArgumentReason.from_bools(
                requires_hyphenation=(
                    "." in argument.value[:-1]
                    or " " in argument.value[:-1]
                    or "_" in argument.value[:-1]
                ),
                requires_lowercasing=bool(re.search(r"[A-Z]", argument.value)),
            )
        )
        if reason is False:
            return

        self.problems[(argument.lineno, argument.col_offset + 1)] = {
            "function_type": function_type,
            "incorrect_name": argument.value,
            "invalid_argument_reason": reason,
        }

    def _check_all_arguments(
        self, decorator_node: ast.Call, function_type: _FunctionType
    ) -> None:
        if decorator_node.args:
            self._check_single_argument(decorator_node.args[0], function_type)

        keyword_argument: ast.keyword
        for keyword_argument in decorator_node.keywords:
            if keyword_argument.arg == "name":
                self._check_single_argument(keyword_argument.value, function_type)
                return

    def _check_decorator(self, decorator_node: ast.expr) -> None:
        if not isinstance(decorator_node, ast.Call):
            return

        if utils.function_call_is_pycord_slash_command_decorator(decorator_node):
            self._check_all_arguments(decorator_node, self._FunctionType.COMMAND)
            return

        if utils.function_call_is_pycord_option_decorator(decorator_node):
            self._check_all_arguments(decorator_node, self._FunctionType.OPTION)
            return

        possible_slash_command_group_name: str
        possible_pycord_decorator_name: str
        match decorator_node.func:
            case ast.Attribute(
                value=ast.Name(id=possible_slash_command_group_name),
                attr=possible_pycord_decorator_name,
            ):
                if (
                    possible_slash_command_group_name
                    in self.plugin.found_slash_command_group_names
                    and possible_pycord_decorator_name
                    in utils.PYCORD_SLASH_COMMAND_DECORATOR_NAMES
                ):
                    self._check_all_arguments(decorator_node, self._FunctionType.COMMAND)
                    return

                if (
                    possible_slash_command_group_name
                    in self.plugin.found_slash_command_group_names
                    and possible_pycord_decorator_name in utils.PYCORD_OPTION_DECORATOR_NAMES
                ):
                    self._check_all_arguments(decorator_node, self._FunctionType.OPTION)
                    return

    @utils.generic_visit_before_return
    @override
    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        decorator_node: ast.expr
        for decorator_node in node.decorator_list:
            self._check_decorator(decorator_node)



"""Linting rule to ensure only imports and the module docstring are above the export."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR105",)


class RuleCAR105(CarrotRule):
    """Linting rule to ensure only imports and the module docstring are above the export."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        line: object | None = ctx.get("line", None)
        if line is not None:
            if not isinstance(line, str):
                raise TypeError

            line = line.strip("\n\r\t `")

        return f"{
            f'`{line if len(line) < 30 else f"{line[:30]}..."}` is not'
            if line
            else "Only import statements & the module's docstring are"
        } allowed above `__all__` export"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        if self.plugin.first_all_export_line_numbers is None or len(tree.body) < 2:
            return

        node: ast.stmt
        for node in tree.body:
            if node.lineno >= self.plugin.first_all_export_line_numbers[0]:
                return

            match node:
                case ast.Expr(value=ast.Constant(value=str())) if node == tree.body[0]:
                    continue
                case (
                    ast.ImportFrom()
                    | ast.Import()
                    | ast.If(test=ast.Name(id="TYPE_CHECKING"), orelse=[])
                    | ast.If(
                        test=ast.Compare(),
                        body=(ast.ImportFrom() | ast.Import(),),
                        orelse=(ast.ImportFrom() | ast.Import(),),
                    )
                ):
                    continue

            self.problems[(node.lineno, 0)] = {"line": ast.unparse(node).split("\n")[0]}



"""Linting rule to ensure the `__all__` export is not missing from a module."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR101",)


class RuleCAR101(CarrotRule):
    """Linting rule to ensure the `__all__` export is not missing from a module."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return "Missing `__all__` export at the top of the module"

    @classmethod
    def get_error_position(
        cls, tree_body: Sequence[ast.stmt], lines: Sequence[str]
    ) -> tuple[int, int]:  # NOTE: I'm sorry to whoever has to work out what is going on here
        """Retrieve the correct error position for this rule based on the problematic AST."""
        if not tree_body:
            return 1, 0

        first_child_node: ast.stmt = tree_body[0]
        first_child_end_lineno: int = first_child_node.end_lineno or first_child_node.lineno

        match first_child_node:
            case ast.Expr(value=ast.Constant()):
                pass
            case _:
                return 1, 0

        if len(tree_body) < 2:
            if not lines[first_child_end_lineno - 1].endswith("\n"):
                return (
                    first_child_end_lineno,
                    (first_child_node.end_col_offset or first_child_node.col_offset),
                )

            return (
                min(
                    first_child_end_lineno + 2,
                    len(lines[first_child_end_lineno:]) + first_child_end_lineno + 1,
                ),
                0,
            )

        second_child_node: ast.stmt = tree_body[1]
        second_child_end_lineno: int = second_child_node.end_lineno or second_child_node.lineno

        match second_child_node:
            case ast.ImportFrom(module="collections.abc", names=[ast.alias(name="Sequence")]):
                pass
            case _:
                if len(lines[first_child_end_lineno : second_child_end_lineno - 1]) in (1, 2):
                    return first_child_end_lineno + 1, 0

                return (
                    min(
                        first_child_end_lineno + 2,
                        (
                            len(lines[first_child_end_lineno : second_child_end_lineno - 1])
                            + first_child_end_lineno
                            + 1
                        ),
                    ),
                    0,
                )

        if len(tree_body) < 3:
            if not lines[second_child_end_lineno - 1].endswith("\n"):
                return (
                    second_child_end_lineno,
                    second_child_node.end_col_offset or second_child_node.col_offset,
                )

            return (
                min(
                    second_child_end_lineno + 2,
                    len(lines[second_child_end_lineno:]) + second_child_end_lineno + 1,
                ),
                0,
            )

        third_child_node: ast.stmt = tree_body[2]
        third_child_end_lineno: int = third_child_node.end_lineno or third_child_node.lineno

        if len(lines[second_child_end_lineno : third_child_end_lineno - 1]) in (1, 2):
            return second_child_end_lineno + 1, 0

        return (
            min(
                second_child_end_lineno + 2,
                (
                    len(lines[second_child_end_lineno : third_child_end_lineno - 1])
                    + second_child_end_lineno
                    + 1
                ),
            ),
            0,
        )

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        if self.plugin.first_all_export_line_numbers is not None:
            return

        error_line_number: int
        error_column_number: int
        error_line_number, error_column_number = self.get_error_position(tree.body, lines)

        self.problems.add_without_ctx(
            (max(error_line_number, 1), max(error_column_number, 0)),
        )



"""Linting rule to ensure preamble lines are separated by only single newlines."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR111",)


class RuleCAR111(CarrotRule):
    """Linting rule to ensure preamble lines are separated by only single newlines."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return (
            "Preamble lines (imports, `__all__` declaration, module docstring, etc.) "
            "should be separated by a single newline"
        )

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        if len(lines) <= 1 or len(tree.body) < 2:
            return

        match tree.body[0]:
            case (
                ast.Expr(value=ast.Constant(value=str()))
                | ast.ImportFrom(module="collections", names=[ast.alias(name="abc.Sequence")])
                | ast.ImportFrom(module="collections", names=[ast.alias(name="abc.Iterable")])
                | ast.ImportFrom(module="collections.abc", names=[ast.alias(name="Sequence")])
                | ast.ImportFrom(module="collections.abc", names=[ast.alias(name="Iterable")])
                | ast.ImportFrom(module="typing", names=[ast.alias(name="Sequence")])
                | ast.ImportFrom(module="typing", names=[ast.alias(name="Iterable")])
                | ast.Import(names=[ast.alias(name="collections")])
                | ast.Import(names=[ast.alias(name="collections.abc")])
                | ast.Import(names=[ast.alias(name="collections.abc.Sequence")])
                | ast.Import(names=[ast.alias(name="collections.abc.Iterable")])
            ):
                pass
            case _:
                return

        match tree.body[1]:
            case (
                ast.ImportFrom(
                    module="collections",
                    names=[ast.alias(name="abc.Sequence"), *_],
                )
                | ast.ImportFrom(
                    module="collections",
                    names=[ast.alias(name="abc.Iterable"), *_],
                )
                | ast.ImportFrom(
                    module="collections.abc",
                    names=[ast.alias(name="Sequence"), *_],
                )
                | ast.ImportFrom(
                    module="collections.abc",
                    names=[ast.alias(name="Iterable"), *_],
                )
                | ast.ImportFrom(module="typing", names=[ast.alias(name="Sequence"), *_])
                | ast.ImportFrom(module="typing", names=[ast.alias(name="Iterable"), *_])
                | ast.Import(names=[ast.alias(name="collections")])
                | ast.Import(names=[ast.alias(name="collections.abc")])
                | ast.Import(names=[ast.alias(name="collections.abc.Sequence")])
                | ast.Import(names=[ast.alias(name="collections.abc.Iterable")])
                | ast.AnnAssign(target=ast.Name(id="__all__"))
                | ast.Assign(targets=[ast.Name(id="__all__"), *_])
            ):
                pass
            case _:
                return

        first_line_end_index: int = (tree.body[0].end_lineno or tree.body[0].lineno) - 1

        if lines[first_line_end_index + 1].strip():
            self.problems.add_without_ctx((first_line_end_index + 1 + 1, 0))
        elif (
            len(lines[first_line_end_index:]) >= 4
            and not lines[first_line_end_index + 2].strip()
        ):
            self.problems.add_without_ctx((first_line_end_index + 2 + 1, 0))

        if not len(tree.body) >= 3:
            return

        match tree.body[2]:
            case (
                ast.AnnAssign(target=ast.Name(id="__all__"))
                | ast.Assign(targets=[ast.Name(id="__all__"), *_])
            ):
                pass
            case _:
                return

        second_line_end_index: int = (tree.body[1].end_lineno or tree.body[1].lineno) - 1

        if lines[second_line_end_index + 1].strip():
            self.problems.add_without_ctx((second_line_end_index + 1 + 1, 0))
        elif (
            len(lines[second_line_end_index:]) >= 4
            and not lines[second_line_end_index + 2].strip()
        ):
            self.problems.add_without_ctx((second_line_end_index + 2 + 1, 0))



"""Linting rule to warn on the unnecessary use of string strip functions."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR140",)


class RuleCAR140(CarrotRule, ast.NodeVisitor):
    """Linting rule to warn on the unnecessary use of string strip functions."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return "Unnecessary use of string strip function"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @utils.generic_visit_before_return
    @override
    def visit_Compare(self, node: ast.Compare) -> None:
        string_object: ast.expr
        fallback_column_number: int
        match node:
            case ast.Compare(
                ops=[ast.In()],
                comparators=[
                    ast.Call(
                        func=ast.Attribute(
                            attr=("strip" | "lstrip" | "rstrip"),
                            value=string_object,
                        ),
                        col_offset=fallback_column_number,
                    ),
                ],
            ):
                self.problems.add_without_ctx(
                    (
                        string_object.end_lineno or string_object.lineno,
                        (
                            string_object.end_col_offset + 1
                            if string_object.end_col_offset is not None
                            else fallback_column_number
                        ),
                    ),
                )
                return

            case _:
                return



"""Linting rule to ensure assignment of `logging.Logger` objects are annotated as final."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR201",)


class RuleCAR201(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure assignment of `logging.Logger` objects are annotated as final."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return "Assignment of `logging.Logger` object should be annotated as `Final[Logger]`"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    def _add_unannotated_problem(self, node: ast.Assign) -> None:
        self.problems.add_without_ctx(
            (
                (
                    node.targets[-1].end_lineno or node.targets[-1].lineno
                    if len(node.targets) > 0
                    else node.lineno
                ),
                (
                    node.targets[-1].end_col_offset or node.targets[-1].col_offset
                    if len(node.targets) > 0
                    else node.col_offset
                ),
            ),
        )

    @classmethod
    def _check_slice_elements(cls, slice_elements: Sequence[ast.expr]) -> bool:
        slice_element: ast.expr
        for slice_element in slice_elements:
            variable_name: str
            match slice_element:
                case ast.Name(id=variable_name):
                    if "logger" in variable_name.lower():
                        return True
                case ast.Constant(value=str(variable_name)):
                    if "logger" in variable_name.lower():
                        return True

        return False

    @classmethod
    def _match_single_target(cls, target: ast.Name | ast.expr) -> bool:
        variable_name: str
        match target:
            case ast.Subscript(value=ast.Name(id=variable_name)):
                if "loggers" in variable_name.lower():
                    return True

        slice_elements: Sequence[ast.expr]
        match target:
            case ast.Name(id=variable_name) if "logger" in variable_name.lower():
                return True

            case ast.Subscript(slice=ast.Name(id=variable_name)) if (
                "logger" in variable_name.lower()
            ):
                return True

            case ast.Subscript(slice=ast.Constant(value=str(variable_name))) if (
                "logger" in variable_name.lower()
            ):
                return True

            case ast.Subscript(slice=ast.Tuple(elts=slice_elements)) if (
                cls._check_slice_elements(slice_elements)
            ):
                return True

            case _:
                return False

    @utils.generic_visit_before_return
    @override
    def visit_Assign(self, node: ast.Assign) -> None:
        match node.value:
            case (
                ast.Call(func=ast.Attribute(value=ast.Name(id="logging"), attr="getLogger"))
                | ast.Call(func=ast.Name(id="getLogger"))
            ):
                self._add_unannotated_problem(node)
                return

        target: ast.expr
        for target in node.targets:
            if self._match_single_target(target):
                self._add_unannotated_problem(node)
                return

    @utils.generic_visit_before_return
    @override
    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        match node.annotation:
            case (
                ast.Constant(value="Final[Logger]")
                | ast.Subscript(value=ast.Name(id="Final"), slice=ast.Name(id="Logger"))
            ):
                return

            case (
                ast.Constant(value=("Logger" | "logging.Logger" | "Final[logging.Logger]"))
                | ast.Name(id="Logger")
                | ast.Attribute(value=ast.Name(id="logging"), attr="Logger")
                | ast.Subscript(
                    value=ast.Name(id="Final"),
                    slice=ast.Attribute(value=ast.Name(id="logging"), attr="Logger"),
                )
            ):
                self.problems.add_without_ctx(
                    (node.annotation.lineno, node.annotation.col_offset),
                )
                return

        if self._match_single_target(node.target):
            self.problems.add_without_ctx(
                (node.annotation.lineno, node.annotation.col_offset),
            )
            return

        match node.value:
            case ast.Call(
                func=(
                    ast.Name(id="getLogger")
                    | ast.Attribute(value=ast.Name(id="logging"), attr="getLogger")
                ),
            ):
                self.problems.add_without_ctx(
                    (node.annotation.lineno, node.annotation.col_offset),
                )
                return

    @override
    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        return

    @override
    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        return

    @override
    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        return

    @override
    def visit_For(self, node: ast.For) -> None:
        return



"""Linting rule to warn when`__init__()` methods are not marked with `@override`."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR163",)


class RuleCAR163(CarrotRule, ast.NodeVisitor):
    """Linting rule to warn when`__init__()` methods are not marked with `@override`."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return "`__init__()` method not marked with `@override`"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @classmethod
    def _has_override_decorator(cls, decorators: Iterable[ast.expr]) -> bool:
        decorator_node: ast.expr
        for decorator_node in decorators:
            match decorator_node:
                case (
                    ast.Name(id="override")
                    | ast.Attribute(value=ast.Name(id="typing"), attr="override")
                    | ast.Attribute(value=ast.Name(id="typing_extensions"), attr="override")
                    | ast.Call(
                        func=(
                            ast.Name(id="override")
                            | ast.Attribute(value=ast.Name(id="typing"), attr="override")
                            | ast.Attribute(
                                value=ast.Name(id="typing_extensions"), attr="override"
                            )
                        )
                    )
                ):
                    return True

        return False

    @utils.generic_visit_before_return
    @override
    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        if node.name != "__init__":
            return

        if not self._has_override_decorator(node.decorator_list):
            self.problems.add_without_ctx((node.lineno, node.col_offset))



"""Linting rule to ensure `logging.Logger` variables contain the word 'logger'."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR202",)


class RuleCAR202(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure `logging.Logger` variables contain the word 'logger'."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return "`logging.Logger` variable name should contain the word 'logger'"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @utils.generic_visit_before_return
    @override
    def visit_Assign(self, node: ast.Assign) -> None:
        if "logger" in "".join(ast.unparse(target) for target in node.targets).lower():
            return

        match node.value:
            case ast.Call(
                func=(
                    ast.Name(id="getLogger")
                    | ast.Attribute(value=ast.Name(id="logging"), attr="getLogger")
                ),
            ):
                self.problems.add_without_ctx((node.lineno, node.col_offset))

    @utils.generic_visit_before_return
    @override
    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        if "logger" in ast.unparse(node.target).lower():
            return

        match node.value:
            case ast.Call(
                func=(
                    ast.Name(id="getLogger")
                    | ast.Attribute(value=ast.Name(id="logging"), attr="getLogger")
                ),
            ):
                self.problems.add_without_ctx((node.target.lineno, node.target.col_offset))

        match node.annotation:
            case (
                ast.Constant(
                    value=(
                        "Logger" | "logging.Logger" | "FinalLogger]" | "Final[logging.Logger]"
                    ),
                )
                | ast.Name(id="Logger")
                | ast.Attribute(value=ast.Name(id="logging"), attr="Logger")
                | ast.Subscript(
                    value=ast.Name(id="Final"),
                    slice=(
                        ast.Name(id="Logger")
                        | ast.Attribute(value=ast.Name(id="logging"), attr="Logger")
                    ),
                )
            ):
                self.problems.add_without_ctx((node.target.lineno, node.target.col_offset))



"""Linting rule to ensure the body of abstract methods only contains the docstring."""  # noqa: N999

import ast
import builtins
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR161",)


class RuleCAR161(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure the body of abstract methods only contains the docstring."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        abstract_method_name: object | None = ctx.get("abstract_method_name", None)
        if abstract_method_name is not None:
            if not isinstance(abstract_method_name, str):
                raise TypeError

            abstract_method_name = abstract_method_name.strip("\n\r\t '()")

        return f"Body of abstract method{
            f" '{
                abstract_method_name
                if len(abstract_method_name) < 30
                else f'{abstract_method_name[:30]}...'
            }()'"
            if abstract_method_name
            else ''
        } should only contain the docstring{
            ' or `pass`'
            if abstract_method_name and abstract_method_name.startswith('_')
            else ''
        } "

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @classmethod
    def _has_abstractmethod_decorator(cls, decorators: Iterable[ast.expr]) -> bool:
        decorator_node: ast.expr
        for decorator_node in decorators:
            match decorator_node:
                case (
                    ast.Name(id="abstractmethod")
                    | ast.Attribute(value=ast.Name(id="abc"), attr="abstractmethod")
                    | ast.Call(
                        func=(
                            ast.Name(id="abstractmethod")
                            | ast.Attribute(value=ast.Name(id="abc"), attr="abstractmethod")
                        )
                    )
                ):
                    return True

        return False

    @classmethod
    def _has_property_decorator(cls, decorators: Iterable[ast.expr]) -> bool:
        decorator_node: ast.expr
        for decorator_node in decorators:
            match decorator_node:
                case (
                    ast.Name(id="property")
                    | ast.Name("cached_property")
                    | ast.Name(id="classproperty")
                    | ast.Name(id="cached_classproperty")
                    | ast.Attribute(value=ast.Name(id="classproperties"), attr="classproperty")
                    | ast.Attribute(
                        value=ast.Name(id="classproperties"), attr="cached_classproperty"
                    )
                    | ast.Call(
                        func=(
                            ast.Name(id="property")
                            | ast.Name("cached_property")
                            | ast.Name(id="classproperty")
                            | ast.Name(id="cached_classproperty")
                            | ast.Attribute(
                                value=ast.Name(id="classproperties"), attr="classproperty"
                            )
                            | ast.Attribute(
                                value=ast.Name(id="classproperties"),
                                attr="cached_classproperty",
                            )
                        )
                    )
                ):
                    return True

        return False

    @classmethod
    def _has_overload_decorator(cls, decorators: Iterable[ast.expr]) -> bool:
        decorator_node: ast.expr
        for decorator_node in decorators:
            match decorator_node:
                case (
                    ast.Name(id="overload")
                    | ast.Attribute(value=ast.Name(id="typing"), attr="overload")
                    | ast.Attribute(value=ast.Name(id="typing_extensions"), attr="overload")
                    | ast.Call(
                        func=(
                            ast.Name(id="overload")
                            | ast.Attribute(value=ast.Name(id="typing"), attr="overload")
                            | ast.Attribute(
                                value=ast.Name(id="typing_extensions"), attr="overload"
                            )
                        )
                    )
                ):
                    return True

        return False

    def _check_function(self, node: ast.FunctionDef | ast.AsyncFunctionDef) -> None:
        if not self._has_abstractmethod_decorator(node.decorator_list):
            return

        body_node: ast.stmt
        for body_node in node.body:
            match body_node:
                case ast.Expr(value=ast.Constant(value=str())):
                    continue
                case ast.Pass() if node.name.strip().startswith(
                    "_"
                ) or self._has_property_decorator(node.decorator_list):
                    continue
                case ast.Expr(value=ast.Constant(value=builtins.Ellipsis)) if (
                    self._has_overload_decorator(node.decorator_list)
                ):
                    continue
                case _:
                    self.problems[(body_node.lineno, body_node.col_offset)] = {
                        "abstract_method_name": node.name
                    }

    @utils.generic_visit_before_return
    @override
    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self._check_function(node)

    @utils.generic_visit_before_return
    @override
    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self._check_function(node)



"""Linting rule to ensure Pycord context command names are capitalised."""  # noqa: N999

import ast
from enum import Enum
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo
    from typing import Final, Self

__all__: Sequence[str] = ("RuleCAR305",)


class _ContextCommandType(Enum):
    MESSAGE = "message"
    USER = "user"

    @classmethod
    def format_value(cls, instance: Self | None) -> str:
        """Produce the formatted string value for this context-command type."""
        return f"{instance.value.strip('\n\r\t -')}-" if instance is not None else ""

    @classmethod
    def get_from_decorator_node(cls, decorator_node: ast.Call) -> _ContextCommandType:
        unparsed_decorator_node: str = ast.unparse(decorator_node).lower()

        if "message" in unparsed_decorator_node:
            return cls.MESSAGE

        if "user" in unparsed_decorator_node:
            return cls.USER

        NON_CONTEXT_COMMAND_MESSAGE: Final[str] = "Decorator node was not a context command."
        raise ValueError(NON_CONTEXT_COMMAND_MESSAGE)


class RuleCAR305(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure Pycord context command names are capitalised."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        context_command_type: object | None = ctx.get("context_command_type", None)
        if context_command_type is not None and not isinstance(
            context_command_type, _ContextCommandType
        ):
            raise TypeError

        return (
            f"Pycord {_ContextCommandType.format_value(context_command_type)}"
            "context command name should be capitalized"
        )

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    def _check_single_argument(
        self, argument: ast.expr, context_command_type: _ContextCommandType
    ) -> None:
        if not isinstance(argument, ast.Constant):
            return

        if not argument.value or not isinstance(argument.value, str):
            return

        if not argument.value[0].isupper():
            self.problems[(argument.lineno, argument.col_offset)] = {
                "context_command_type": context_command_type,
            }

    def _check_all_arguments(self, decorator_node: ast.Call) -> None:
        context_command_type: _ContextCommandType = (
            _ContextCommandType.get_from_decorator_node(decorator_node)
        )

        if decorator_node.args:
            self._check_single_argument(decorator_node.args[0], context_command_type)

        keyword_argument: ast.keyword
        for keyword_argument in decorator_node.keywords:
            if keyword_argument.arg == "name":
                self._check_single_argument(keyword_argument.value, context_command_type)
                return

    def _check_decorator(self, decorator_node: ast.expr) -> None:
        if not isinstance(decorator_node, ast.Call):
            return

        if utils.function_call_is_pycord_context_command_decorator(decorator_node):
            self._check_all_arguments(decorator_node)
            return

        possible_slash_command_group_name: str
        possible_pycord_decorator_name: str
        match decorator_node.func:
            case ast.Attribute(
                value=ast.Name(id=possible_slash_command_group_name),
                attr=possible_pycord_decorator_name,
            ):
                if (
                    possible_slash_command_group_name
                    in self.plugin.found_slash_command_group_names
                    and possible_pycord_decorator_name
                    in utils.PYCORD_CONTEXT_COMMAND_DECORATOR_NAMES
                ):
                    self._check_all_arguments(decorator_node)
                    return

    @utils.generic_visit_before_return
    @override
    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        decorator_node: ast.expr
        for decorator_node in node.decorator_list:
            self._check_decorator(decorator_node)



"""Linting rule to warn about declaring `*args` or `**kwargs` as function parameters."""  # noqa: N999

import ast
from enum import Enum
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR150",)


class RuleCAR150(CarrotRule, ast.NodeVisitor):
    """Linting rule to warn about declaring `*args` or `**kwargs` as function parameters."""

    class _InvalidArgumentType(Enum):
        STAR_ARGS = "`*args`"
        STAR_STAR_KWARGS = "`**kwargs`"

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        invalid_argument_type: object | None = ctx.get("invalid_argument_type", None)
        if invalid_argument_type is not None and not isinstance(
            invalid_argument_type, cls._InvalidArgumentType
        ):
            raise TypeError

        return f"Use of {
            invalid_argument_type.value
            if invalid_argument_type is not None
            else '`*args` or `**kwargs`'
        } in function definition"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    def _check_args(self, args: ast.arguments) -> None:
        if args.vararg is not None:
            self.problems[(args.vararg.lineno, args.vararg.col_offset - 1)] = {
                "invalid_argument_type": self._InvalidArgumentType.STAR_ARGS,
            }

        if args.kwarg is not None:
            self.problems[(args.kwarg.lineno, args.kwarg.col_offset - 2)] = {
                "invalid_argument_type": self._InvalidArgumentType.STAR_STAR_KWARGS,
            }

    @utils.generic_visit_before_return
    @override
    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self._check_args(node.args)

    @utils.generic_visit_before_return
    @override
    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self._check_args(node.args)



"""Linting rule to prevent the use of dataclasses."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR501",)


class RuleCAR501(CarrotRule, ast.NodeVisitor):
    """Linting rule to prevent the use of dataclasses."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return 'Use of dataclass found, declare class manually without "magic" instead'

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @utils.generic_visit_before_return
    @override
    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        decorator_node: ast.expr
        for decorator_node in node.decorator_list:
            decorator_name: str
            match decorator_node:
                case ast.Name(id=decorator_name) | ast.Call(func=ast.Name(id=decorator_name)):
                    if "dataclass" in decorator_name.lower():
                        self.problems.add_without_ctx(
                            (decorator_node.lineno, decorator_node.col_offset),
                        )
                        continue

            decorator_module: ast.expr
            match decorator_node:
                case (
                    ast.Attribute(value=decorator_module, attr=decorator_name)
                    | ast.Call(
                        func=ast.Attribute(value=decorator_module, attr=decorator_name),
                    )
                ):
                    if "dataclass" in (ast.unparse(decorator_module) + decorator_name).lower():
                        self.problems.add_without_ctx(
                            (decorator_node.lineno, decorator_node.col_offset),
                        )
                        continue



"""Linting rule class-property names should be in all caps."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR162",)


class RuleCAR162(CarrotRule, ast.NodeVisitor):
    """Linting rule class-property names should be in all caps."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        classproperty_name: object | None = ctx.get("classproperty_name", None)
        if classproperty_name is not None:
            if not isinstance(classproperty_name, str):
                raise TypeError

            classproperty_name = classproperty_name.strip("\n\r\t '")

        return f"Class-property name{
            f" '{classproperty_name}'" if classproperty_name else ''
        } should be in all caps{
            f": '{classproperty_name.upper()}'" if classproperty_name else ''
        }"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @classmethod
    def _has_classproperty_decorator(cls, decorators: Iterable[ast.expr]) -> bool:
        decorator_node: ast.expr
        for decorator_node in decorators:
            match decorator_node:
                case (
                    ast.Name(id="classproperty")
                    | ast.Name(id="cached_classproperty")
                    | ast.Attribute(value=ast.Name(id="classproperties"), attr="classproperty")
                    | ast.Attribute(
                        value=ast.Name(id="classproperties"), attr="cached_classproperty"
                    )
                    | ast.Call(
                        func=(
                            ast.Name(id="classproperty")
                            | ast.Name(id="cached_classproperty")
                            | ast.Attribute(
                                value=ast.Name(id="classproperties"), attr="classproperty"
                            )
                            | ast.Attribute(
                                value=ast.Name(id="classproperties"),
                                attr="cached_classproperty",
                            )
                        )
                    )
                ):
                    return True

        return False

    def _check_function(self, node: ast.FunctionDef | ast.AsyncFunctionDef) -> None:
        if not self._has_classproperty_decorator(node.decorator_list):
            return

        if not node.name.isupper():
            self.problems[(node.lineno, node.col_offset)] = {"classproperty_name": node.name}

    @utils.generic_visit_before_return
    @override
    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self._check_function(node)

    @utils.generic_visit_before_return
    @override
    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self._check_function(node)



"""Linting rules under the "TXB" category."""

from typing import TYPE_CHECKING, override

from typed_classproperties import classproperty

from flake8_carrot.utils import BasePlugin, TeXBotRule

if TYPE_CHECKING:
    from collections.abc import Collection, Sequence

__all__: Sequence[str] = ("TeXBotPlugin", "TeXBotRule")


class TeXBotPlugin(BasePlugin):
    """Plugin class holding all "TXB" rules to be run on some code provided by Flake8."""

    @classproperty
    @override
    def RULES(cls) -> Collection[type[TeXBotRule]]:
        return frozenset()



"""Linting rule to ensure all uses of `astpretty.pprint` are removed."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR401",)


class RuleCAR401(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure all uses of `astpretty.pprint` are removed."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return "`astpretty.pprint` found"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @utils.generic_visit_before_return
    @override
    def visit_Call(self, node: ast.Call) -> None:
        function_name: str
        line_number: int
        column_number: int
        match node.func:
            case ast.Name(id=function_name, lineno=line_number, col_offset=column_number):
                if not self.plugin.pprint_imported_for_debugging:
                    return

                if "print" in function_name.lower():
                    self.problems.add_without_ctx((line_number, column_number))
                    return

                return

        module_name: str
        match node.func:
            case ast.Attribute(
                value=ast.Name(id=module_name),
                attr=function_name,
                lineno=line_number,
                col_offset=column_number,
            ):
                if module_name in utils.PPRINT_MODULES and "print" in function_name.lower():
                    self.problems.add_without_ctx((line_number, column_number))
                    return

                return



"""Linting rule to ensure `re.fullmatch()` is used over `re.match()`."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR601",)


class RuleCAR601(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure `re.fullmatch()` is used over `re.match()`."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return "Prefer to use `re.fullmatch()` over `re.match()`"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @utils.generic_visit_before_return
    @override
    def visit_Call(self, node: ast.Call) -> None:
        match node:
            case ast.Call(
                func=(
                    ast.Attribute(value=ast.Name(id="re"), attr="match") | ast.Name(id="match")
                ),
                args=[ast.Constant(value=str()), *_],
            ):
                self.problems.add_without_ctx((node.func.lineno, node.func.col_offset))
                return

        first_argument: str
        match node:
            case ast.Call(
                func=(
                    ast.Attribute(value=ast.Name(id="re"), attr="match") | ast.Name(id="match")
                ),
                args=[ast.Name(id=first_argument), *_],
            ):
                if first_argument.isupper():
                    self.problems.add_without_ctx((node.func.lineno, node.func.col_offset))
                    return



"""
Linting rule to ensure `re.fullmatch()` is used over `re.search()`.

Only applicable when the regex pattern uses beginning and ending line anchors.
"""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR602",)


class RuleCAR602(CarrotRule, ast.NodeVisitor):
    """
    Linting rule to ensure `re.fullmatch()` is used over `re.search()`.

    Only applicable when the regex pattern uses beginning and ending line anchors.
    """

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return (
            "Use `re.fullmatch()` over `re.search()` "
            "when using beginning & ending line anchors"
        )

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @classmethod
    def _look_for_single_re_multiline(cls, arg: ast.expr) -> bool:
        match arg:
            case ast.Attribute(value=ast.Name(id="re"), attr="MULTILINE"):
                return True

        left: ast.expr
        right: ast.expr
        match arg:
            case ast.BinOp(op=ast.BitOr(), left=left, right=right):
                return cls._look_for_single_re_multiline(
                    left
                ) or cls._look_for_single_re_multiline(right)

        return False

    @utils.generic_visit_before_return
    @override
    def visit_Call(self, node: ast.Call) -> None:
        regex: str
        remaining_args: Iterable[ast.expr]
        match node:
            case ast.Call(
                func=(
                    ast.Attribute(value=ast.Name(id="re"), attr="search")
                    | ast.Name(id="search")
                ),
                args=[ast.Constant(value=str(regex)), *remaining_args],
            ):
                if regex.startswith(r"\A") and regex.endswith(r"\Z"):
                    self.problems.add_without_ctx((node.func.lineno, node.func.col_offset))
                    return

                if (
                    any(self._look_for_single_re_multiline(arg) for arg in remaining_args)
                    and regex.startswith(r"^")
                    and regex.endswith(r"$")
                ):
                    self.problems.add_without_ctx((node.func.lineno, node.func.col_offset))
                    return

                return



"""Test suite to check the functionality of all CAR plugin rules."""

import abc
import itertools
import re
from typing import TYPE_CHECKING

import pytest

from flake8_carrot import CarrotPlugin
from flake8_carrot.utils import CarrotRule
from tests._testing_utils import apply_plugin_to_ast

if TYPE_CHECKING:
    from collections.abc import Sequence
    from collections.abc import Set as AbstractSet

    from flake8_carrot.utils import CarrotRule

__all__: Sequence[str] = ("TestRuleCAR001",)


class BaseTestCarrotPlugin(abc.ABC):  # noqa: B024
    @classmethod
    def _apply_carrot_plugin_to_ast(cls, raw_testing_ast: str) -> AbstractSet[str]:
        return apply_plugin_to_ast(raw_testing_ast, CarrotPlugin)


class TestRuleMessages(BaseTestCarrotPlugin):
    """Test suite to ensure all CAR rules have consistent message formatting."""

    @pytest.mark.parametrize("rule_class", CarrotPlugin.RULES)
    def test_message_never_ends_with_full_stop(self, rule_class: type[CarrotRule]) -> None:
        """Ensure each rule message is formatted correctly without a trailing full-stop."""
        assert not rule_class.format_error_message(ctx={}).endswith(".")

    @pytest.mark.parametrize("rule_class", CarrotPlugin.RULES)
    def test_no_double_zero_in_rule_code(self, rule_class: type[CarrotRule]) -> None:
        """Ensure each rule code is indexed correctly without double zeros."""
        assert "00" not in rule_class.__name__

    @pytest.mark.parametrize("rule_class", CarrotPlugin.RULES)
    def test_all_rule_classes_named_with_rule_code(self, rule_class: type[CarrotRule]) -> None:
        """Ensure each message contains some sort of rule code."""
        assert re.fullmatch(r"\ARuleCAR[0-9]{1,3}\Z", rule_class.__name__)

    @pytest.mark.parametrize("rule_class", CarrotPlugin.RULES)
    def test_correct_rule_code_in_error_message(self, rule_class: type[CarrotRule]) -> None:
        """Ensure each message contains the correct corresponding rule code."""
        rule_number_match: re.Match[str] | None = re.fullmatch(
            r"\A(?:Rule)?CAR([0-9]{1,3})\Z", rule_class.__name__, re.IGNORECASE
        )

        assert re.fullmatch(
            (
                rf"\ACAR{
                    rule_number_match.group(1)
                    if rule_number_match is not None
                    else r'[0-9]{1,3}'
                } .+\Z"
            ),
            rule_class.format_error_message(ctx={}),
        )

    @pytest.mark.parametrize("rule_class", CarrotPlugin.RULES)
    def test_no_double_rule_code_in_error_message(self, rule_class: type[CarrotRule]) -> None:
        """Ensure rule messages don't statically contain the rule code (twice)."""
        assert not re.fullmatch(
            r"\A\s*CAR\s*[0-9]{1,4}.*\Z",
            rule_class._format_error_message(ctx={}),  # noqa: SLF001
        )
        assert not re.fullmatch(
            r"\A\s*CAR\s*[0-9]{1,4}\s*CAR\s*[0-9]{1,4}.*\Z",
            rule_class.format_error_message(ctx={}),
        )


class TestRuleCAR001(BaseTestCarrotPlugin):
    """Test suite for rule CAR001."""

    @classmethod
    def _get_message(cls, line_number: int, column_number: int) -> str:
        return (
            f"{line_number}:{column_number} CAR101 "
            "Missing `__all__` export at the top of the module"
        )

    @pytest.mark.parametrize(
        ("raw_test_ast", "expected_error_position"),
        (
            ("", (1, 1)),
            (" ", (1, 1)),
            ("\n", (1, 1)),
            ("class Foo:\n    pass", (1, 1)),
            ("\nclass Foo:\n    pass", (1, 1)),
            ("class Foo:\n    pass\n", (1, 1)),
            ('"""This is a docstring."""', (1, 27)),
            ('"""This is a docstring."""\n', (2, 1)),
            *(
                (f'"""This is a docstring."""\n{"\n" * count}', (3, 1))
                for count in range(1, 5)
            ),
            ('"""\nThis is a docstring.\nwow!\n"""', (4, 4)),
            ('"""\nThis is a docstring.\nwow!\n"""\n', (5, 1)),
            *(
                (f'"""\nThis is a docstring.\nwow!\n"""\n{"\n" * count}', (6, 1))
                for count in range(1, 5)
            ),
            *(
                (f'"""This is a docstring."""{"\n" * count}x=3\n', (2, 1))
                for count in range(1, 4)
            ),
            *(
                (f'"""This is a docstring."""\n\n\n{"\n" * count}x=3\n', (3, 1))
                for count in range(1, 4)
            ),
            *(
                (f'"""\nThis is a docstring.\nwow!\n"""{"\n" * count}x=3\n', (5, 1))
                for count in range(1, 4)
            ),
            *(
                (f'"""\nThis is a docstring.\nwow!\n"""\n\n\n{"\n" * count}x=3\n', (6, 1))
                for count in range(1, 4)
            ),
            ('"""This is a docstring."""\n\nfrom collections.abc import Sequence', (3, 37)),
            ('"""This is a docstring."""\n\nfrom collections.abc import Sequence\n', (4, 1)),
            *(
                (
                    (
                        '"""This is a docstring."""\n\n'
                        f"from collections.abc import Sequence\n{'\n' * count}"
                    ),
                    (5, 1),
                )
                for count in range(1, 5)
            ),
            (
                '"""\nThis is a docstring.\nwow!\n"""\n\nfrom collections.abc import Sequence',
                (6, 37),
            ),
            *(
                (
                    (
                        '"""\nThis is a docstring.\nwow!\n"""\n\n'
                        f"from collections.abc import Sequence\n{'\n' * count}"
                    ),
                    (8, 1),
                )
                for count in range(1, 5)
            ),
            *(
                (
                    (
                        '"""This is a docstring."""\n\n'
                        f"from collections.abc import Sequence{'\n' * count}x=3\n"
                    ),
                    (4, 1),
                )
                for count in range(1, 4)
            ),
            *(
                (
                    (
                        '"""This is a docstring."""\n\n'
                        f"from collections.abc import Sequence\n\n\n{'\n' * count}x=3\n"
                    ),
                    (5, 1),
                )
                for count in range(1, 4)
            ),
            *(
                (
                    (
                        '"""\nThis is a docstring.\nwow!\n"""\n\n'
                        f"from collections.abc import Sequence{'\n' * count}x=3\n"
                    ),
                    (7, 1),
                )
                for count in range(1, 4)
            ),
            *(
                (
                    (
                        '"""\nThis is a docstring.\nwow!\n"""\n\n'
                        f"from collections.abc import Sequence\n\n\n{'\n' * count}x=3\n"
                    ),
                    (8, 1),
                )
                for count in range(1, 3)
            ),
        ),
    )
    def test_missing_all_export(
        self, raw_test_ast: str, expected_error_position: tuple[int, int]
    ) -> None:
        """Ensure rule CAR001 is alerted for modules without __all__ exports."""
        assert self._get_message(*expected_error_position) in self._apply_carrot_plugin_to_ast(
            raw_test_ast
        )


class TestRuleCAR101(BaseTestCarrotPlugin):
    """Test suite for rule CAR101."""

    @pytest.mark.parametrize(
        "raw_test_ast",
        (
            "__all__ = ()",
            '"""This is a docstring."""\n__all__ = ()\n',
            (
                '"""This is a docstring."""\n'
                "from collections.abc import Sequence\n__all__: Sequence[str] = ()\n"
            ),
            'from collections.abc import Sequence\nx=3\n__all__: Sequence[str] = ("x",)\n',
        ),
    )
    def test_successful_all_export_provided(self, raw_test_ast: str) -> None:
        """Ensure rule CAR101 is not alerted for modules with __all__ exports."""
        assert all(
            "CAR101" not in problem
            for problem in self._apply_carrot_plugin_to_ast(raw_test_ast)
        )


class TestRuleCAR111(BaseTestCarrotPlugin):
    """Test suite for rule CAR111."""

    @classmethod
    def _get_message(cls, line_number: int, column_number: int) -> str:
        return (
            f"{line_number}:{column_number} CAR111 "
            "Preamble lines (imports, `__all__` declaration, module docstring, etc.) "
            "should be separated by a single newline"
        )

    @pytest.mark.parametrize(
        "raw_test_ast",
        (
            "",
            "  ",
            *("\n" * count for count in range(1, 6)),
            "  \n",
            (
                "#! /usr/bin/env python3\n"
                '"""This is a docstring."""\n\n'
                "from collections.abc import Sequence\n\n"
                "__all__: Sequence[str] = ()\n\n"
                'print("hello")\n'
            ),
            (
                "#! /usr/bin/env python3\n\n"
                '"""This is a docstring."""\n\n'
                "from collections.abc import Sequence\n\n"
                "__all__: Sequence[str] = ()\n\n"
                'print("hello")\n'
            ),
            (
                '"""This is a docstring."""\n\n'
                "from collections.abc import Sequence\n\n"
                "__all__: Sequence[str] = ()\n\n"
                'print("hello")\n'
            ),
            (
                '"""This is a docstring."""\n\n'
                "from collections.abc import Sequence\n\n"
                '__all__: Sequence[str] = ("foo",)\n\n'
                "def foo():\n    "
                'print("hello")\n'
            ),
            (
                "\n"
                '"""This is a docstring."""\n\n'
                "from collections.abc import Sequence\n\n"
                '__all__: Sequence[str] = ("foo",)\n\n'
                "def foo():\n    "
                'print("hello")\n'
            ),
            (
                "\n\n"
                '"""This is a docstring."""\n\n'
                "from collections.abc import Sequence\n\n"
                '__all__: Sequence[str] = ("foo",)\n\n'
                "def foo():\n    "
                'print("hello")\n'
            ),
            (
                '"""This is a docstring."""\n\n'
                "from collections.abc import Sequence\n\n"
                '__all__: Sequence[str] = ("foo",)\n\n'
                "import random\n\n"
                "def foo():\n    "
                'print(f"hello {random.random()}")\n'
            ),
            (
                "\n"
                '"""This is a docstring."""\n\n'
                "from collections.abc import Sequence\n\n"
                '__all__: Sequence[str] = ("foo",)\n\n'
                "import random\n\n"
                "def foo():\n    "
                'print(f"hello {random.random()}")\n'
            ),
            (
                '"""This is a docstring."""\n\n'
                "from collections.abc import Sequence\n\n"
                '__all__: Sequence[str] = ("foo",)\n\n'
                "import random\n"
            ),
            '"""This is a docstring."""\n\nimport random\n',
            '\n"""This is a docstring."""\n\nimport random\n',
            *(
                f'{"\n" * count_pre}"""This is a docstring."""{"\n" * count_post}'
                for count_post in range(0, 5)
                for count_pre in range(0, 5)
            ),
            *(
                f"{'\n' * count_pre}from collections.abc import Sequence{'\n' * count_post}"
                for count_post in range(0, 5)
                for count_pre in range(0, 5)
            ),
            *(
                f"{'\n' * count_pre}__all__: Sequence[str] = (){'\n' * count_post}"
                for count_post in range(0, 5)
                for count_pre in range(0, 5)
            ),
            *(
                f"{'\n' * count_pre}__all__ = (){'\n' * count_post}"
                for count_post in range(0, 5)
                for count_pre in range(0, 5)
            ),
            *(
                f"{'\n' * count_pre}x = 5{'\n' * count_post}"
                for count_post in range(0, 5)
                for count_pre in range(0, 5)
            ),
            *(
                f"{'\n' * count_pre}import random{'\n' * count_post}"
                for count_post in range(0, 5)
                for count_pre in range(0, 5)
            ),
            "import random\nx = random.random()\n",
            "import random\nx = random.random()\n\nx = 5\n",
            '"""This is a docstring."""\nx = 4\n\n\ny = 5\n',
        ),
    )
    def test_successful_single_newline_between_preamble(self, raw_test_ast: str) -> None:
        """Ensure rule CAR111 is alerted for incorrect lines between preamble blocks."""
        problems: AbstractSet[str] = self._apply_carrot_plugin_to_ast(raw_test_ast)
        assert not problems or any("CAR111" not in problem for problem in problems)

    @pytest.mark.parametrize(
        ("raw_test_ast", "expected_error_position"),
        (
            *(
                (
                    f"{'\n' * count}{line1}\n{line2}\n",
                    (2 + count, 1),
                    # NOTE: Test for missing single newline (first break)
                )
                for line1, line2 in itertools.combinations(
                    (
                        '"""This is a docstring."""',
                        "from collections.abc import Sequence",
                        "__all__ = ()",
                    ),
                    2,
                )
                for count in range(0, 4)
            ),
            *(
                (
                    (
                        f"{'\n' * count_pre}"
                        f'"""This is a docstring."""{"\n" * count_post}'
                        "from collections.abc import Sequence\n"
                    ),
                    (3 + count_pre, 1),
                    # NOTE: Test that always second newline is flagged when > 1 newlines (first break)
                )
                for count_post in range(3, 6)
                for count_pre in range(0, 4)
            ),
            *(
                (
                    (
                        f"{'\n' * count_pre}"
                        f'"""This is a docstring."""{"\n" * count_post1}'
                        f"from collections.abc import Sequence{'\n' * count_post2}"
                        "__all__: Sequence[str] = ()\n"
                    ),
                    (3 + count_pre, 1),
                    # NOTE: Test that always second newline is flagged when > 1 newlines (first break) and incorrect newlines below
                )
                for count_post1 in range(3, 6)
                for count_pre in range(0, 4)
                for count_post2 in range(1, 6)
                if count_post2 != 2
            ),
            *(
                (
                    (
                        f'{"\n" * count}"""This is a docstring."""\n\n'
                        "from collections.abc import Sequence\n"
                        "__all__: Sequence[str] = ()\n"
                    ),
                    (4 + count, 1),
                    # NOTE: Test for missing single newline (second break)
                )
                for count in range(0, 4)
            ),
            *(
                (
                    (
                        f'{"\n" * count_pre}"""This is a docstring."""\n\n'
                        f"from collections.abc import Sequence{'\n' * count_post}"
                        "__all__: Sequence[str] = ()\n"
                    ),
                    (5 + count_pre, 1),
                    # NOTE: Test that always second newline is flagged when > 1 newlines (second break)
                )
                for count_pre in range(0, 4)
                for count_post in range(3, 6)
            ),
            *(
                (
                    (
                        f"{'\n' * count_pre}"
                        f'"""This is a docstring."""{"\n" * count_post1}'
                        f"from collections.abc import Sequence{'\n' * count_post2}"
                        "__all__: Sequence[str] = ()\n"
                    ),
                    (3 + count_pre + count_post1, 1),
                    # NOTE: Test that always second newline is flagged when > 1 newlines (second break) and newlines above
                )
                for count_post1 in range(1, 6)
                for count_pre in range(0, 4)
                for count_post2 in range(3, 6)
                if count_post1 != 2
            ),
        ),
    )
    def test_no_multiple_newlines_between_preamble(
        self, raw_test_ast: str, expected_error_position: tuple[int, int]
    ) -> None:
        assert self._get_message(*expected_error_position) in self._apply_carrot_plugin_to_ast(
            raw_test_ast
        )


class TestRuleCAR610(BaseTestCarrotPlugin):
    """Test suite for rule CAR610."""

    @classmethod
    def _get_message(cls, line_number: int, column_number: int) -> str:
        return (
            f"{line_number}:{column_number} CAR610 "
            'Regex pattern string should use a raw string: `r"..."`'
        )

    @pytest.mark.parametrize(
        "raw_test_ast",
        (
            're.search(r"\\A\\w+[a-z]\\Z", content)',
            "re.search(rf\"\\A<(?:{\n'|'.join(url_schemes)})>\\Z\", content)",
            're.match(r"\\A\\w+[a-z]\\Z", content)',
            "re.match(rf\"\\A<(?:{\n'|'.join(url_schemes)})>\\Z\", content)",
            're.fullmatch(r"\\A\\w+[a-z]\\Z", content)',
            "re.fullmatch(rf\"\\A<(?:{\n'|'.join(url_schemes)})>\\Z\", content)",
        ),
    )
    def test_correctly_using_r_prefix(self, raw_test_ast: str) -> None:
        """Ensure rule CAR610 is not alerted for modules with an `rf` string prefix."""
        assert all(
            "CAR610" not in problem
            for problem in self._apply_carrot_plugin_to_ast(raw_test_ast)
        )

    @pytest.mark.parametrize(
        ("raw_test_ast", "expected_error_position"),
        (
            ('re.search("\\\\A\\\\w+[a-z]\\\\Z", content)', (1, 1)),
            ("re.search(f\"\\\\A<(?:{\n'|'.join(url_schemes)})>\\\\Z\", content)", (1, 1)),
            ('re.match("\\\\A\\\\w+[a-z]\\\\Z", content)', (1, 1)),
            ("re.match(f\"\\\\A<(?:{\n'|'.join(url_schemes)})>\\\\Z\", content)", (1, 1)),
            ('re.fullmatch("\\\\A\\\\w+[a-z]\\\\Z", content)', (1, 1)),
            ("re.fullmatch(f\"\\\\A<(?:{\n'|'.join(url_schemes)})>\\\\Z\", content)", (1, 1)),
        ),
    )
    def test_missing_prefix(
        self, raw_test_ast: str, expected_error_position: tuple[int, int]
    ) -> None:
        """Ensure rule CAR610 is not alerted for modules with an `rf` string prefix."""
        assert self._get_message(*expected_error_position) in self._apply_carrot_plugin_to_ast(
            raw_test_ast
        )



"""Linting rule to suggest removing IDE specific ignore comments."""  # noqa: N999

import re
import tokenize
from typing import TYPE_CHECKING, override

from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    import ast
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR124",)


class RuleCAR124(CarrotRule):
    """Linting rule to suggest removing IDE specific ignore comments."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return "IDE specific ignore comments (E.g. `# noinspection ...`) should be removed"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        file_token: TokenInfo
        for file_token in file_tokens:
            if file_token.type != tokenize.COMMENT:
                continue

            match: re.Match[str] | None = re.fullmatch(
                r"\A.*(#\s+noinspection).*\Z", file_token.string
            )
            if match is not None:
                self.problems.add_without_ctx(
                    (file_token.start[0], file_token.start[1] + match.span(1)[0]),
                )
                continue



"""Custom opinionated linting rules to adhere Python code to CarrotManMatt's style guide."""

from typing import TYPE_CHECKING

from .carrot import CarrotPlugin
from .tex_bot import TeXBotPlugin

if TYPE_CHECKING:
    from collections.abc import Sequence

__all__: Sequence[str] = ("CarrotPlugin", "TeXBotPlugin")



"""Linting rule to ensure only a singular `__all__` export is present."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR102",)


class RuleCAR102(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure only a singular `__all__` export is present."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return "Multiple `__all__` exports found in a single module"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @classmethod
    def _get_all_assignment(cls, targets: Iterable[ast.expr]) -> ast.Name | None:
        target: ast.expr
        for target in targets:
            match target:
                case ast.Name(id="__all__"):
                    return target
                case ast.Tuple():
                    return cls._get_all_assignment(target.elts)

        return None

    @utils.generic_visit_before_return
    @override
    def visit_Assign(self, node: ast.Assign) -> None:
        if self.plugin.first_all_export_line_numbers is None:
            return

        all_assignment: ast.Name | None = self._get_all_assignment(node.targets)
        if all_assignment is None:
            return

        if all_assignment.lineno == self.plugin.first_all_export_line_numbers[0]:
            return

        self.problems.add_without_ctx((all_assignment.lineno, all_assignment.col_offset))

    @utils.generic_visit_before_return
    @override
    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        if self.plugin.first_all_export_line_numbers is None:
            return

        if node.lineno == self.plugin.first_all_export_line_numbers[0]:
            return

        match node.target:
            case ast.Name(id="__all__"):
                self.problems.add_without_ctx((node.lineno, node.col_offset))



"""Test suite to check the functionality of all TXB plugin rules."""

from typing import TYPE_CHECKING

from flake8_carrot import TeXBotPlugin
from tests._testing_utils import apply_plugin_to_ast

if TYPE_CHECKING:
    from collections.abc import Sequence
    from collections.abc import Set as AbstractSet

__all__: Sequence[str] = ()


def _apply_tex_bot_plugin_to_ast(raw_testing_ast: str) -> AbstractSet[str]:
    return apply_plugin_to_ast(raw_testing_ast, TeXBotPlugin)



"""Linting rule to ensure regex patterns use raw strings."""  # noqa: N999

import ast
import re
import tokenize
from io import StringIO
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Collection, Iterable, Mapping, Sequence
    from tokenize import TokenInfo
    from typing import Final

    from flake8_carrot.carrot import CarrotPlugin

__all__: Sequence[str] = ("RuleCAR610",)


class RuleCAR610(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure regex patterns use raw strings."""

    @override
    def __init__(self, plugin: CarrotPlugin) -> None:
        self.source: str | None = None

        super().__init__(plugin)

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        return 'Regex pattern string should use a raw string: `r"..."`'

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.source = "".join(lines)
        self.visit(tree)

    def _check_node_for_incorrect_string_type(
        self, argument: ast.Constant | ast.JoinedStr
    ) -> None:
        TOKENS: Final[Iterable[TokenInfo]] = tokenize.generate_tokens(
            StringIO(
                (
                    ast.get_source_segment(self.source, argument)
                    if self.source is not None
                    else ast.unparse(argument)
                ),
            ).readline,
        )

        token: TokenInfo
        for token in TOKENS:
            if (
                (isinstance(argument, ast.Constant) and token.type == tokenize.STRING)
                or token.type == tokenize.FSTRING_START
            ) and not re.search(r"\A(?:rf?|fr)[\"']", token.string):
                self.problems.add_without_ctx(
                    (argument.lineno - 1 + token.start[0], token.start[1]),
                )

    @utils.generic_visit_before_return
    @override
    def visit_Call(self, node: ast.Call) -> None:
        RE_FUNCTION_NAMES: Final[Collection[str]] = (
            "search",
            "match",
            "fullmatch",
            "findall",
            "finditer",
            "compile",
            "split",
            "sub",
            "subn",
        )

        function_name: str
        argument: ast.Constant | ast.JoinedStr
        match node:
            case ast.Call(
                func=(
                    ast.Attribute(value=ast.Name(id="re"), attr=function_name)
                    | ast.Name(id=function_name)
                ),
                args=[(ast.Constant(value=str()) | ast.JoinedStr()) as argument, *_],
            ):
                if function_name not in RE_FUNCTION_NAMES:
                    return

                self._check_node_for_incorrect_string_type(argument)
                return



"""Linting rule to ensure classes are not defined inside functions."""  # noqa: N999

import ast
from typing import TYPE_CHECKING, override

from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR160",)


class RuleCAR160(CarrotRule, ast.NodeVisitor):
    """Linting rule to ensure classes are not defined inside functions."""

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        function_name: object | None = ctx.get("function_name", None)
        if function_name is not None:
            if not isinstance(function_name, str):
                raise TypeError

            function_name = function_name.strip("\n\r\t '()")

        class_name: object | None = ctx.get("class_name", None)
        if class_name is not None:
            if not isinstance(class_name, str):
                raise TypeError

            class_name = class_name.strip("\n\r\t '")

        is_function_async: object | None = ctx.get("is_function_async", None)
        if class_name is not None and not isinstance(is_function_async, bool):
            raise TypeError

        return f"Class{
            f" '{class_name if len(class_name) < 30 else f'{class_name[:30]}...'}'"
            if class_name
            else ''
        } defined inside {'async ' if is_function_async else ''}function{
            f" '{function_name if len(function_name) < 30 else f'{function_name[:30]}...'}()'"
            if function_name
            else ''
        }"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    def _check_function(self, node: ast.FunctionDef | ast.AsyncFunctionDef) -> None:
        inner_node: ast.AST
        for inner_node in ast.walk(node):
            if isinstance(inner_node, ast.ClassDef):
                self.problems[(inner_node.lineno, inner_node.col_offset)] = {
                    "function_name": node.name,
                    "class_name": inner_node.name,
                    "is_function_async": isinstance(node, ast.AsyncFunctionDef),
                }

    # NOTE: Don't descend into visitor children
    @override
    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self._check_function(node)

    # NOTE: Don't descend into visitor children
    @override
    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self._check_function(node)



"""Linting rule to suggest replacing repeated boolean operators with `all()`/`any()`."""  # noqa: N999

import ast
from enum import Enum
from typing import TYPE_CHECKING, override

from flake8_carrot import utils
from flake8_carrot.utils import CarrotRule

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from tokenize import TokenInfo

__all__: Sequence[str] = ("RuleCAR180",)


class RuleCAR180(CarrotRule, ast.NodeVisitor):
    """Linting rule to suggest replacing repeated boolean operators with `all()`/`any()`."""

    class _OpType(Enum):
        OR = "or", "any(...)"
        AND = "and", "all(...)"

    @classmethod
    @override
    def _format_error_message(cls, ctx: Mapping[str, object]) -> str:
        op_type: object | None = ctx.get("op_type", None)
        if op_type is not None and not isinstance(op_type, cls._OpType):
            raise TypeError

        return f"Replace repeated {
            f' `... {op_type.value[0]} ...`' if op_type is not None else 'boolean operator'
        } with {f' `{op_type.value[1]}`' if op_type is not None else 'call to `any()`/'}"

    @override
    def run_check(
        self, tree: ast.Module, file_tokens: Sequence[TokenInfo], lines: Sequence[str]
    ) -> None:
        self.visit(tree)

    @utils.generic_visit_before_return
    @override
    def visit_BoolOp(self, node: ast.BoolOp) -> None:
        if len(node.values) < 4:
            return

        match node.op:
            case ast.Or():
                self.problems[(node.lineno, node.col_offset)] = {"op_type": self._OpType.OR}
            case ast.And():
                self.problems[(node.lineno, node.col_offset)] = {"op_type": self._OpType.AND}
